#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td

from holidays.calendars import _HebrewLunisolar
from holidays.holiday_base import HolidayBase


class Israel(HolidayBase):
    country = "IL"

    def _add_with_eve(self, name: str, dt: date) -> None:
        self._add_holiday(f"{name} - Eve", dt + td(days=-1))
        self._add_holiday(name, dt)

    def _populate(self, year):
        if year <= 1947:
            return None

        if year >= 2101:
            raise NotImplementedError

        super()._populate(year)

        # Passover
        passover_dt = _HebrewLunisolar.hebrew_holiday_date(year, "PASSOVER")
        self._add_with_eve("Passover I", passover_dt)
        for offset in range(1, 5):
            self._add_holiday("Passover - Chol HaMoed", passover_dt + td(days=offset))
        self._add_with_eve("Passover VII", passover_dt + td(days=+6))

        # Memorial Day
        memorial_day_dt = _HebrewLunisolar.hebrew_holiday_date(year, "MEMORIAL_DAY")
        observed_delta = 0
        if self.observed:
            if self._is_thursday(memorial_day_dt):
                observed_delta = -1
            elif self._is_friday(memorial_day_dt):
                observed_delta = -2
            elif year >= 2004 and self._is_sunday(memorial_day_dt):
                observed_delta = 1

        name = "Memorial Day"
        if observed_delta != 0:
            self._add_holiday(f"{name} (Observed)", memorial_day_dt + td(days=observed_delta))
        else:
            self._add_holiday(name, memorial_day_dt)

        # Independence Day
        name = "Independence Day"
        if self.observed and observed_delta != 0:
            self._add_holiday(f"{name} (Observed)", memorial_day_dt + td(days=observed_delta + 1))
        else:
            self._add_holiday(name, memorial_day_dt + td(days=+1))

        # Lag Baomer
        lag_baomer_dt = _HebrewLunisolar.hebrew_holiday_date(year, "LAG_BAOMER")
        self._add_holiday("Lag B'Omer", lag_baomer_dt)

        # Shavuot
        shavuot_dt = _HebrewLunisolar.hebrew_holiday_date(year, "SHAVUOT")
        self._add_with_eve("Shavuot", shavuot_dt)

        # Rosh Hashana
        rosh_hashanah_dt = _HebrewLunisolar.hebrew_holiday_date(year, "ROSH_HASHANAH")
        name = "Rosh Hashanah"
        self._add_with_eve(name, rosh_hashanah_dt)
        self._add_holiday(name, rosh_hashanah_dt + td(days=+1))

        # Yom Kippur
        yom_kippur_dt = _HebrewLunisolar.hebrew_holiday_date(year, "YOM_KIPPUR")
        self._add_with_eve("Yom Kippur", yom_kippur_dt)

        # Sukkot
        sukkot_dt = _HebrewLunisolar.hebrew_holiday_date(year, "SUKKOT")
        self._add_with_eve("Sukkot I", sukkot_dt)
        for offset in range(1, 6):
            self._add_holiday("Sukkot - Chol HaMoed", sukkot_dt + td(days=offset))
        self._add_with_eve("Sukkot VII", sukkot_dt + td(days=+7))

        # Hanukkah
        # Some o prior's year Hannukah may fall in current year.
        for yr in (year - 1, year):
            hanukkah_dt = _HebrewLunisolar.hebrew_holiday_date(yr, "HANUKKAH")
            for offset in range(8):
                self._add_holiday("Hanukkah", hanukkah_dt + td(days=offset))

        # Purim
        purim_dt = _HebrewLunisolar.hebrew_holiday_date(year, "PURIM")
        self._add_with_eve("Purim", purim_dt)
        self._add_holiday("Shushan Purim", purim_dt + td(days=+1))


class IL(Israel):
    pass


class ISR(Israel):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from typing import Tuple, Union

from holidays.calendars.gregorian import MAR, APR, MAY, JUN, JUL, AUG, SEP, NOV, DEC, MON
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class UnitedKingdom(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_the_United_Kingdom
    """

    country = "GB"
    special_holidays = {
        1977: (JUN, 7, "Silver Jubilee of Elizabeth II"),
        1981: (JUL, 29, "Wedding of Charles and Diana"),
        1999: (DEC, 31, "Millennium Celebrations"),
        2002: (JUN, 3, "Golden Jubilee of Elizabeth II"),
        2011: (APR, 29, "Wedding of William and Catherine"),
        2012: (JUN, 5, "Diamond Jubilee of Elizabeth II"),
        2022: (
            (JUN, 3, "Platinum Jubilee of Elizabeth II"),
            (SEP, 19, "State Funeral of Queen Elizabeth II"),
        ),
        2023: (MAY, 8, "Coronation of Charles III"),
    }
    subdivisions: Union[Tuple[()], Tuple[str, ...]] = (
        "ENG",  # England
        "NIR",  # Northern Ireland
        "SCT",  # Scotland
        "WLS",  # Wales
    )
    _deprecated_subdivisions: Tuple[str, ...] = (
        "England",
        "Northern Ireland",
        "Scotland",
        "UK",
        "Wales",
    )

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_observed(self, dt: date, days: int = +1) -> None:
        if self.observed and self._is_weekend(dt):
            self._add_holiday(
                "%s (Observed)" % self[dt], dt + td(days=+2 if self._is_saturday(dt) else days)
            )

    def _populate(self, year: int) -> None:
        super()._populate(year)

        # Good Friday
        self._add_good_friday("Good Friday")

        # May Day bank holiday (first Monday in May)
        if year >= 1978:
            dt = (
                date(year, MAY, 8)
                # In 2020 moved to Friday to mark 75th anniversary of VE Day.
                if year in {1995, 2020}
                else self._get_nth_weekday_of_month(1, MON, MAY)
            )
            self._add_holiday("May Day", dt)

        # Spring bank holiday (last Monday in May)
        if year >= 1971:
            spring_bank_dates = {
                2002: date(year, JUN, 4),
                2012: date(year, JUN, 4),
                2022: date(year, JUN, 2),
            }
            dt = spring_bank_dates.get(year, self._get_nth_weekday_of_month(-1, MON, MAY))
            self._add_holiday("Spring Bank Holiday", dt)

        if self.subdiv == "England":
            self._add_subdiv_eng_holidays()
        elif self.subdiv == "Northern Ireland":
            self._add_subdiv_nir_holidays()
        elif self.subdiv == "Scotland":
            self._add_subdiv_sct_holidays()
        elif self.subdiv == "Wales":
            self._add_subdiv_wls_holidays()

    def _add_subdiv_holidays(self):
        if self.subdiv not in {"SCT", "Scotland"}:
            # New Year's Day
            if self._year >= 1975:
                self._add_observed(self._add_new_years_day("New Year's Day"))

            # Christmas Day
            self._add_observed(self._add_christmas_day("Christmas Day"), days=+2)

            # Boxing Day
            self._add_observed(self._add_christmas_day_two("Boxing Day"), days=+2)

        super()._add_subdiv_holidays()

    def _add_subdiv_eng_holidays(self):
        # Easter Monday
        self._add_easter_monday("Easter Monday")

        # Late Summer bank holiday (last Monday in August)
        if self._year >= 1971:
            self._add_holiday(
                "Late Summer Bank Holiday", self._get_nth_weekday_of_month(-1, MON, AUG)
            )

    def _add_subdiv_nir_holidays(self):
        if self._year >= 1903:
            # St. Patrick's Day
            self._add_observed(self._add_holiday("St. Patrick's Day", MAR, 17))

        # Easter Monday
        self._add_easter_monday("Easter Monday")

        # Battle of the Boyne
        self._add_holiday("Battle of the Boyne", JUL, 12)

        # Late Summer bank holiday (last Monday in August)
        if self._year >= 1971:
            self._add_holiday(
                "Late Summer Bank Holiday", self._get_nth_weekday_of_month(-1, MON, AUG)
            )

    def _add_subdiv_sct_holidays(self):
        # New Year's Day
        name = "New Year's Day"
        jan_1 = self._add_new_years_day(name)
        if self.observed and self._is_weekend(jan_1):
            self._add_holiday(
                "%s (Observed)" % name, jan_1 + td(days=+3 if self._is_saturday(jan_1) else +1)
            )

        # New Year Holiday
        name = "New Year Holiday"
        jan_2 = self._add_new_years_day_two(name)
        self._add_observed(jan_2)
        if self.observed and self._is_monday(jan_2):
            self._add_new_years_day_three("%s (Observed)" % name)

        # Summer bank holiday (first Monday in August)
        self._add_holiday("Summer Bank Holiday", self._get_nth_weekday_of_month(1, MON, AUG))

        if self._year >= 2006:
            # St. Andrew's Day
            self._add_holiday("St. Andrew's Day", NOV, 30)

        # Christmas Day
        self._add_observed(
            self._add_christmas_day("Christmas Day"), days=+2 if self._year >= 1974 else +1
        )

        if self._year >= 1974:
            # Boxing Day
            self._add_observed(self._add_christmas_day_two("Boxing Day"), days=+2)

    def _add_subdiv_wls_holidays(self):
        # Easter Monday
        self._add_easter_monday("Easter Monday")

        # Late Summer bank holiday (last Monday in August)
        if self._year >= 1971:
            self._add_holiday(
                "Late Summer Bank Holiday", self._get_nth_weekday_of_month(-1, MON, AUG)
            )


class UK(UnitedKingdom):
    pass


class GB(UnitedKingdom):
    pass


class GBR(UnitedKingdom):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td

from holidays.calendars.gregorian import FEB, APR, JUN, AUG, SEP, OCT, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Kenya(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Kenya
    http://kenyaembassyberlin.de/Public-Holidays-in-Kenya.48.0.html
    https://www.officeholidays.com/holidays/kenya/moi-day
    """

    country = "KE"
    special_holidays = {
        2020: (FEB, 11, "President Moi Celebration of Life Day"),
        2022: (
            (APR, 29, "State Funeral for Former President Mwai Kibaki"),
            (AUG, 9, "Election Day"),
            (SEP, 10, "Day of Mourning for Queen Elizabeth II"),
            (SEP, 11, "Day of Mourning for Queen Elizabeth II"),
            (SEP, 12, "Day of Mourning for Queen Elizabeth II"),
            (SEP, 13, "Inauguration Day"),
        ),
    }

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        def _add_observed(dt: date, days: int = +1) -> None:
            if self.observed and self._is_sunday(dt):
                self._add_holiday("%s (Observed)" % self[dt], dt + td(days=days))

        if year <= 1962:
            return None

        super()._populate(year)

        # New Year's Day
        _add_observed(self._add_new_years_day("New Year's Day"))

        # Good Friday
        self._add_good_friday("Good Friday")

        # Easter Monday
        self._add_easter_monday("Easter Monday")

        # Labour Day
        _add_observed(self._add_labor_day("Labour Day"))

        if year >= 2010:
            # Mandaraka Day
            _add_observed(self._add_holiday("Madaraka Day", JUN, 1))

        if 2002 <= year <= 2009 or year >= 2018:
            _add_observed(
                # Utamaduni/Moi Day
                self._add_holiday("Utamaduni Day" if year >= 2021 else "Moi Day", OCT, 10)
            )

        _add_observed(
            # Mashuja/Kenyatta Day
            self._add_holiday("Mashujaa Day" if year >= 2010 else "Kenyatta Day", OCT, 20)
        )

        # Jamhuri Day
        _add_observed(self._add_holiday("Jamhuri Day", DEC, 12))

        # Christmas Day
        _add_observed(self._add_christmas_day("Christmas Day"), days=+2)

        # Boxing Day
        _add_observed(self._add_christmas_day_two("Boxing Day"))


class KE(Kenya):
    pass


class KEN(Kenya):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)


def _normalize_tuple(data):
    """Normalize tuple.

    :return:
        An unchanged object for tuple of tuples, e.g., ((JAN, 10), (DEC, 31)).
        An object put into a tuple otherwise, e.g., ((JAN, 10),).
    """
    return data if not data or type(data[0]) == tuple else (data,)



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.calendars.gregorian import MAR, APR, MAY, AUG, OCT, NOV
from holidays.calendars.julian import JULIAN_CALENDAR
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Georgia(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    Georgia holidays.

    References:
     - https://en.wikipedia.org/wiki/Public_holidays_in_Georgia_(country)
    """

    country = "GE"
    default_language = "ka"
    supported_languages = ("en_US", "ka")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self, JULIAN_CALENDAR)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("ახალი წელი"))
        self._add_new_years_day_two(tr("ბედობა"))

        # Christmas Day (Orthodox).
        self._add_christmas_day(tr("ქრისტეშობა"))

        # Baptism Day of our Lord Jesus Christ.
        self._add_epiphany_day(tr("ნათლისღება"))

        # Mother's Day.
        self._add_holiday(tr("დედის დღე"), MAR, 3)

        # Women's Day,
        self._add_womens_day(tr("ქალთა საერთაშორისო დღე"))

        # Easter.
        self._add_good_friday(tr("წითელი პარასკევი"))
        self._add_holy_saturday(tr("დიდი შაბათი"))
        self._add_easter_sunday(tr("აღდგომა"))
        self._add_easter_monday(tr("შავი ორშაბათი"))

        # National Unity Day,
        self._add_holiday(tr("ეროვნული ერთიანობის დღე"), APR, 9)

        # Day of Victory.
        self._add_world_war_two_victory_day(tr("ფაშიზმზე გამარჯვების დღე"))

        # Saint Andrew the First-Called Day.
        self._add_holiday(tr("წმინდა ანდრია პირველწოდებულის დღე"), MAY, 12)

        # Independence Day.
        self._add_holiday(tr("დამოუკიდებლობის დღე"), MAY, 26)

        # Saint Mary's Day.
        self._add_holiday(tr("მარიამობა"), AUG, 28)

        # Day of Svetitskhoveli Cathedral
        self._add_holiday(tr("სვეტიცხოვლობა"), OCT, 14)

        # Saint George's Day
        self._add_holiday(tr("გიორგობა"), NOV, 23)


class GE(Georgia):
    pass


class GEO(Georgia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)


import warnings
from datetime import date
from datetime import timedelta as td

from holidays.calendars import _CustomChineseCalendar
from holidays.calendars.gregorian import JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, SAT, SUN
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import (
    ChineseCalendarHolidays,
    ChristianHolidays,
    InternationalHolidays,
)


class SouthKorea(HolidayBase, ChineseCalendarHolidays, ChristianHolidays, InternationalHolidays):
    """
    1. https://publicholidays.co.kr/ko/2020-dates/
    2. https://en.wikipedia.org/wiki/Public_holidays_in_South_Korea
    3. https://www.law.go.kr/%EB%B2%95%EB%A0%B9/%EA%B4%80%EA%B3%B5%EC%84%9C%EC%9D%98%20%EA%B3%B5%ED%9C%B4%EC%9D%BC%EC%97%90%20%EA%B4%80%ED%95%9C%20%EA%B7%9C%EC%A0%95  # noqa

    According to (3), the alt holidays in Korea are as follows:
    The alt holiday means next first non holiday after the holiday.
    Independence movement day, Liberation day, National Foundation Day,
    Hangul Day, Children's Day, Birthday of the Buddha, Christmas Day have alt holiday if they fell on Saturday or Sunday.
    Lunar New Year's Day, Korean Mid Autumn Day have alt holiday if they fell
    on only sunday.

    """

    country = "KR"
    special_holidays = {
        # Just for year 2020 - since 2020.08.15 is Sat, the government
        # decided to make 2020.08.17 holiday, yay
        2020: (AUG, 17, "Alternative public holiday"),
    }

    def __init__(self, *args, **kwargs):
        ChineseCalendarHolidays.__init__(self, calendar=SouthKoreaLunisolarCalendar())
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_with_alt_holiday(
        self,
        hol_name: str,
        hol_date: date,
        add_hol: bool = True,
        since: int = 2021,
        include_sat: bool = True,
    ) -> None:
        """Add alternative holiday on first day from the date provided
        that's not already a another holiday nor a weekend.

        :param hol_name:
           The name of the holiday.

        :param hol_date:
           The date of the holiday.

        :param add_hol:
           Add the holiday itself, not alternative only

        :param since:
           Year starting from which alt holiday should be added

        :param include_sat:
           Whether Saturday is to be considered a weekend in addition to
           Sunday.
        """
        if add_hol:
            self._add_holiday(hol_name, hol_date)

        if not self.observed:
            return None

        target_weekday = {SUN}
        if include_sat:
            target_weekday.add(SAT)
        if (
            hol_date.weekday() in target_weekday or len(self.get_list(hol_date)) > 1
        ) and hol_date.year >= since:
            hol_date += td(days=+1)
            while hol_date.weekday() in target_weekday or hol_date in self:
                hol_date += td(days=+1)
            self._add_holiday("Alternative holiday of %s" % hol_name, hol_date)

    def _populate(self, year):
        if year <= 1947:
            return None
        super()._populate(year)

        # New Year's Day
        self._add_new_years_day("New Year's Day")

        preceding_day = "The day preceding of %s"
        second_day = "The second day of %s"

        # Lunar New Year
        name = "Lunar New Year's Day"
        new_year_date = self._add_chinese_new_years_day(name)
        self._add_chinese_new_years_eve(preceding_day % name)
        self._add_chinese_new_years_day_two(second_day % name)

        for delta in (-1, 0, +1):
            self._add_with_alt_holiday(
                name, new_year_date + td(days=delta), add_hol=False, since=2015, include_sat=False
            )

        # Independence Movement Day
        self._add_with_alt_holiday("Independence Movement Day", date(year, MAR, 1))

        # Tree Planting Day
        # removed from holiday since 2006
        if 1949 <= year <= 2005 and year != 1960:
            self._add_holiday("Tree Planting Day", APR, 5)

        # Birthday of the Buddha
        name = "Birthday of the Buddha"
        self._add_with_alt_holiday(
            name, self._add_chinese_birthday_of_buddha(name), add_hol=False, since=2023
        )

        # Children's Day
        if year >= 1975:
            self._add_with_alt_holiday("Children's Day", date(year, MAY, 5), since=2015)

        # Labour Day
        name = "Labour Day"
        if year >= 1994:
            self._add_labor_day(name)
        else:
            self._add_holiday(name, MAR, 10)

        # Memorial Day
        self._add_holiday("Memorial Day", JUN, 6)

        # Constitution Day
        # removed from holiday since 2008
        if year <= 2007:
            self._add_holiday("Constitution Day", JUL, 17)

        # Liberation Day
        self._add_with_alt_holiday("Liberation Day", date(year, AUG, 15))

        # Korean Mid Autumn Day
        name = "Chuseok"
        chuseok_date = self._add_mid_autumn_festival(name)
        self._add_holiday(preceding_day % name, chuseok_date + td(days=-1))
        self._add_holiday(second_day % name, chuseok_date + td(days=+1))

        for delta in (-1, 0, +1):
            self._add_with_alt_holiday(
                name, chuseok_date + td(days=delta), add_hol=False, since=2014, include_sat=False
            )

        # National Foundation Day
        self._add_with_alt_holiday("National Foundation Day", date(year, OCT, 3))

        # Hangeul Day
        if year <= 1990 or year >= 2013:
            self._add_with_alt_holiday("Hangeul Day", date(year, OCT, 9))

        # Christmas Day
        name = "Christmas Day"
        self._add_with_alt_holiday(name, self._add_christmas_day(name), add_hol=False, since=2023)


class Korea(SouthKorea):
    def __init__(self, *args, **kwargs) -> None:
        warnings.warn("Korea is deprecated, use SouthKorea instead.", DeprecationWarning)

        super().__init__(*args, **kwargs)


class KR(SouthKorea):
    pass


class KOR(SouthKorea):
    pass


class SouthKoreaLunisolarCalendar(_CustomChineseCalendar):
    BUDDHA_BIRTHDAY_DATES = {
        1931: (MAY, 25),
        1968: (MAY, 5),
        2001: (MAY, 1),
        2012: (MAY, 28),
        2023: (MAY, 27),
        2025: (MAY, 5),
    }

    LUNAR_NEW_YEAR_DATES = {
        1916: (FEB, 4),
        1944: (JAN, 26),
        1954: (FEB, 4),
        1958: (FEB, 19),
        1966: (JAN, 22),
        1988: (FEB, 18),
        1997: (FEB, 8),
        2027: (FEB, 7),
        2028: (JAN, 27),
    }

    MID_AUTUMN_DATES = {
        1942: (SEP, 25),
        2040: (SEP, 21),
    }



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import JAN, NOV, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Monaco(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Monaco
    https://en.service-public-entreprises.gouv.mc/Employment-and-social-affairs/Employment-regulations/Leave/Public-Holidays  # noqa: E501
    """

    country = "MC"
    default_language = "fr"
    special_holidays = {2015: (JAN, 7, tr("Jour férié"))}
    supported_languages = ("en_US", "fr", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_observed(self, dt: date, days: int = +1) -> None:
        if self.observed and self._is_sunday(dt):
            self._add_holiday(self.tr("%s (Observé)") % self[dt], dt + td(days=days))

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_observed(self._add_new_years_day(tr("Le jour de l'An")))

        # Saint Devote's Day.
        self._add_holiday(tr("La Sainte Dévote"), JAN, 27)

        # Easter Monday.
        self._add_easter_monday(tr("Le lundi de Pâques"))

        # Labour Day.
        self._add_observed(self._add_labor_day(tr("Fête de la Travaille")))

        # Ascension's Day.
        self._add_ascension_thursday(tr("L'Ascension"))

        # Whit Monday.
        self._add_whit_monday(tr("Le lundi de Pentecôte"))

        # Corpus Christi.
        self._add_corpus_christi_day(tr("La Fête Dieu"))

        # Assumption's Day.
        self._add_observed(self._add_assumption_of_mary_day(tr("L'Assomption de Marie")))

        # All Saints' Day.
        self._add_observed(self._add_all_saints_day(tr("La Toussaint")))

        # Prince's Day.
        self._add_observed(self._add_holiday(tr("La Fête du Prince"), NOV, 19))

        dt = (DEC, 8)
        if year >= 2019 and self._is_sunday(*dt):
            dt = (DEC, 9)
        # Immaculate Conception's Day.
        self._add_holiday(tr("L'Immaculée Conception"), *dt)

        # Christmas Day.
        self._add_observed(self._add_christmas_day(tr("Noël")))


class MC(Monaco):
    pass


class MCO(Monaco):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td

from holidays.calendars.gregorian import FEB, MAY, JUN, OCT
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, IslamicHolidays, InternationalHolidays


class Nigeria(HolidayBase, ChristianHolidays, InternationalHolidays, IslamicHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Nigeria
    """

    country = "NG"
    special_holidays = {
        2019: (
            (FEB, 22, "Public Holiday for Elections"),
            (MAY, 29, "Presidential Inauguration Day"),
        ),
    }

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        if year <= 1978:
            return None

        super()._populate(year)
        observed_dates = set()

        # New Year's Day
        observed_dates.add(self._add_new_years_day("New Year's Day"))

        self._add_good_friday("Good Friday")
        self._add_easter_monday("Easter Monday")

        # Worker's day
        if year >= 1981:
            observed_dates.add(self._add_labor_day("Workers' Day"))

        # Democracy day moved around after its inception in 2000
        # Initally it fell on May 29th
        # In 2018 it was announced that the holiday
        # will move to June 12th from 2019
        if year >= 2000:
            observed_dates.add(
                self._add_holiday("Democracy Day", *((JUN, 12) if year >= 2019 else (MAY, 29)))
            )

        # Independence Day
        observed_dates.add(self._add_holiday("Independence Day", OCT, 1))

        # Christmas day
        observed_dates.add(self._add_christmas_day("Christmas Day"))

        # Boxing day
        observed_dates.add(self._add_christmas_day_two("Boxing Day"))

        # Eid al-Fitr - Feast Festive
        # This is an estimate
        # date of observance is announced yearly
        observed_dates.update(self._add_eid_al_fitr_day("Eid-el-Fitr"))
        observed_dates.update(self._add_eid_al_fitr_day_two("Eid-el-Fitr Holiday"))

        # Eid al-Adha - Scarfice Festive
        # This is an estimate
        # date of observance is announced yearly
        observed_dates.update(self._add_eid_al_adha_day("Eid-el-Kabir"))
        observed_dates.update(self._add_eid_al_adha_day_two("Eid-el-Kabir Holiday"))

        # Birthday of Prophet Muhammad
        observed_dates.update(self._add_mawlid_day("Eid-el-Mawlid"))

        # Observed holidays
        if self.observed and year >= 2016:
            for hol_date in sorted(observed_dates):
                if not self._is_weekend(hol_date):
                    continue
                obs_date = hol_date + td(days=+1)
                while self._is_weekend(obs_date) or obs_date in observed_dates:
                    obs_date += td(days=+1)
                for hol_name in self.get_list(hol_date):
                    observed_dates.add(self._add_holiday("%s (Observed)" % hol_name, obs_date))


class NG(Nigeria):
    pass


class NGA(Nigeria):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from holidays.calendars.gregorian import JAN, FEB, MAR, APR, MAY, OCT, NOV
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays


class VaticanCity(HolidayBase, ChristianHolidays):
    """
    References:
      - https://en.wikipedia.org/wiki/Public_holidays_in_Vatican_City
      - https://www.ewtn.com/catholicism/library/solemnity-of-mary-mother-of-god-5826
      - https://www.franciscanmedia.org/saint-of-the-day/saint-joseph-the-worker/
    """

    country = "VA"

    def __init__(self, *args, **kwargs) -> None:
        ChristianHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year: int) -> None:
        if year <= 1928:
            return None
        super()._populate(year)

        # Solemnity of Mary Day.
        # This is supposedly the same as International New Year.
        # Modern adoption across the entire Latin Church in 1931 though this
        # was already celebrated in Rome as the Octave day of Christmas.
        self._add_holiday("Solemnity of Mary Day", JAN, 1)

        # Epiphany.
        self._add_epiphany_day("Epiphany")

        # Lateran Treaty Day.
        self._add_holiday("Lateran Treaty Day", FEB, 11)

        if year >= 1978:
            if year >= 2013:
                # Anniversary of the election of Pope Francis.
                dt = (MAR, 13)
            elif year >= 2005:
                # Anniversary of the election of Pope Benedict XVI.
                dt = (APR, 19)
            else:
                # Anniversary of the election of Pope John Paul II.
                dt = (OCT, 16)
            self._add_holiday("Anniversary of the Election of the Holy Father", *dt)

        # In 2005-2013 - also name day for the civilian name of
        # Pope Benedict XVI (Josef Ratzinger)
        # Saint Joseph's Day.
        self._add_saint_josephs_day("Saint Joseph's Day")

        # Easter Sunday.
        self._add_easter_sunday("Easter Sunday")

        # Easter Monday.
        self._add_easter_monday("Easter Monday")

        if year >= 2013:
            # Name day for the civilian name of Pope Francis
            # (Jorge Mario Bergoglio)
            # Saint George's Day.
            self._add_saint_georges_day("Saint George's Day")

        if year >= 1955:
            # Saint Joseph the Worker's Day.
            # Created in response to May Day holidays by Pope Pius XII in 1955.
            self._add_holiday("Saint Joseph the Worker's Day", MAY, 1)

        if year <= 2009:
            # Ascension of Christ.
            self._add_ascension_thursday("Ascension of Christ")

            # Corpus Christi.
            self._add_corpus_christi_day("Corpus Christi")

        # Saints Peter and Paul.
        self._add_saints_peter_and_paul_day("Saint Peter and Saint Paul's Day")

        # Assumption of Mary Day.
        self._add_assumption_of_mary_day("Assumption Day")

        # Nativity Of Mary Day.
        self._add_nativity_of_mary_day("Nativity of Mary Day")

        # All Saints' Day.
        self._add_all_saints_day("All Saints' Day")

        if 1978 <= year <= 2004:
            # Name day for the civilian name of Pope John Paul II
            # (Karol Józef Wojtyła)
            # Saint Charles Borromeo Day.
            self._add_holiday("Saint Charles Borromeo Day", NOV, 4)

        # Immaculate Conception.
        self._add_immaculate_conception_day("Immaculate Conception Day")

        # Christmas Day.
        self._add_christmas_day("Christmas Day")

        # Saint Stephen's Day.
        self._add_christmas_day_two("Saint Stephen's Day")


class VA(VaticanCity):
    pass


class VAT(VaticanCity):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#           tasnim<tasnimislam1999@gmail.com>
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from holidays.calendars.gregorian import FEB, MAR, APR, AUG, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import InternationalHolidays


class Bangladesh(HolidayBase, InternationalHolidays):
    """
    https://mopa.gov.bd/sites/default/files/files/mopa.gov.bd/public_holiday/61c35b73_e335_462a_9bcf_4695b23b6d82/reg4-2019-212.PDF
    https://en.wikipedia.org/wiki/Public_holidays_in_Bangladesh
    """

    country = "BD"

    def __init__(self, *args, **kwargs):
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # International Mother's language Day.
        self._add_holiday("International Mother's language Day", FEB, 21)

        # Sheikh Mujibur Rahman's Birthday and Children's Day.
        self._add_holiday("Sheikh Mujibur Rahman's Birthday and Children's Day", MAR, 17)

        # Independence Day.
        self._add_holiday("Independence Day", MAR, 26)

        # Bengali New Year's Day.
        self._add_holiday("Bengali New Year's Day", APR, 14)

        # May Day.
        self._add_labor_day("May Day")

        # National Mourning Day.
        self._add_holiday("National Mourning Day", AUG, 15)

        # Victory Day.
        self._add_holiday("Victory Day", DEC, 16)


class BD(Bangladesh):
    pass


class BGD(Bangladesh):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td

from holidays.calendars.gregorian import MAR, MAY, JUL, AUG, OCT, DEC
from holidays.calendars.julian import JULIAN_CALENDAR
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, IslamicHolidays, InternationalHolidays


class Kazakhstan(HolidayBase, ChristianHolidays, InternationalHolidays, IslamicHolidays):
    """
    1. https://www.officeholidays.com/countries/kazakhstan/2020
    2. https://egov.kz/cms/en/articles/holidays-calend
    3. https://en.wikipedia.org/wiki/Public_holidays_in_Kazakhstan
    4. https://adilet.zan.kz/rus/docs/Z010000267_/history
    """

    country = "KZ"

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self, JULIAN_CALENDAR)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        # Kazakhstan declared its sovereignty on 25 October 1990
        if year <= 1990:
            return None

        super()._populate(year)
        observed_dates = set()

        # New Year's holiday (2 days)
        name = "New Year"
        observed_dates.add(self._add_new_years_day(name))
        observed_dates.add(self._add_new_years_day_two(name))

        # Orthodox Christmas (nonworking day, without extending)
        if year >= 2006:
            self._add_christmas_day("Orthodox Christmas")

        # International Women's Day
        observed_dates.add(self._add_womens_day("International Women's Day"))

        # Nauryz holiday
        if year >= 2002:
            name = "Nauryz holiday"
            observed_dates.add(self._add_holiday(name, MAR, 22))
            if year >= 2010:
                observed_dates.add(self._add_holiday(name, MAR, 21))
                observed_dates.add(self._add_holiday(name, MAR, 23))

        # Kazakhstan People Solidarity Holiday
        observed_dates.add(self._add_labor_day("Kazakhstan People Solidarity Holiday"))

        # Defender of the Fatherland Day
        if year >= 2013:
            observed_dates.add(self._add_holiday("Defender of the Fatherland Day", MAY, 7))

        # Victory Day
        observed_dates.add(self._add_world_war_two_victory_day("Victory Day"))

        # Capital Day
        if year >= 2009:
            observed_dates.add(self._add_holiday("Capital Day", JUL, 6))

        # Constitution Day of the Republic of Kazakhstan
        if year >= 1996:
            observed_dates.add(
                self._add_holiday("Constitution Day of the Republic of Kazakhstan", AUG, 30)
            )

        # Republic Day
        if 1994 <= year <= 2008 or year >= 2022:
            observed_dates.add(self._add_holiday("Republic Day", OCT, 25))

        # First President Day
        if 2012 <= year <= 2021:
            observed_dates.add(self._add_holiday("First President Day", DEC, 1))

        # Kazakhstan Independence Day
        name = "Kazakhstan Independence Day"
        observed_dates.add(self._add_holiday(name, DEC, 16))
        if 2002 <= year <= 2021:
            observed_dates.add(self._add_holiday(name, DEC, 17))

        # Kurban Ait (nonworking day, without extending)
        if year >= 2006:
            self._add_eid_al_adha_day("Kurban Ait")

        if self.observed and year >= 2002:
            for hol_date in sorted(observed_dates):
                if not self._is_weekend(hol_date):
                    continue
                obs_date = hol_date + td(days=+1)
                while self._is_weekend(obs_date) or obs_date in observed_dates:
                    obs_date += td(days=+1)
                observed_dates.add(self._add_holiday("%s (Observed)" % self[hol_date], obs_date))


class KZ(Kazakhstan):
    pass


class KAZ(Kazakhstan):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from gettext import gettext as tr

from holidays.calendars.gregorian import (
    JAN,
    FEB,
    MAR,
    APR,
    MAY,
    JUN,
    JUL,
    AUG,
    SEP,
    OCT,
    NOV,
    DEC,
    MON,
)
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Argentina(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    A subclass of :py:class:`HolidayBase` representing public holidays
    in Argentina.


    References:

    - Based on:
        https://es.wikipedia.org/wiki/Anexo:D%C3%ADas_feriados_en_Argentina
    - [Ley 24455] Belgrano and San Martin Day as third Monday
    - [Ley 27399] - For 2018++
        https://www.argentina.gob.ar/normativa/nacional/ley-27399-281835/texto
    - [Decreto 1585/2010] - 2011-2013 Bridge Holidays, Movable Holidays Law
    - [Decreto 1768/2013] - 2014-2016 Bridge Holidays
    - [Decretos 52-80-923/2017] - 2017-2019 Bridge Holidays
    - [Decreto 717/2019] - 2020 Bridge Holidays
    - [Decreto 297/2020] - Veteran Day moved due to Covid-19
    - [Decreto 947/2020] - 2021 Bridge Holidays
    - [Decreto 789/2021] - 2022 Bridge Holidays
    - [Decreto 764/2022] - 2023 Bridge Holidays
    - [Always Update Calendar Year Link]
        https://www.argentina.gob.ar/interior/feriados
        http://servicios.lanacion.com.ar/feriados
        https://www.clarin.com/feriados/
    - [Specific Calendar Year]
        https://www.lanacion.com.ar/feriados/2024/
        https://www.argentina.gob.ar/interior/feriados-nacionales-2023
        https://www.argentina.gob.ar/interior/feriados-nacionales-2022
        https://www.argentina.gob.ar/interior/feriados-nacionales-2021
        https://www.argentina.gob.ar/interior/feriados-nacionales-2020
        https://www.cultura.gob.ar/feriados-2019-en-argentina_7326/
        https://servicios.lanacion.com.ar/app-mobile/feriados/2019
        https://servicios.lanacion.com.ar/app-mobile/feriados/2018
        https://servicios.lanacion.com.ar/app-mobile/feriados/2017
        https://servicios.lanacion.com.ar/app-mobile/feriados/2016
        https://servicios.lanacion.com.ar/app-mobile/feriados/2015
    """

    country = "AR"
    default_language = "es"
    supported_languages = ("en_US", "es", "uk")

    # Special Bridge Holidays are given upto 3 days a year
    # as long as it's declared 50 days before calendar year's end
    # There's no Bridge Holidays declared in 2017

    # Bridge Public Holiday.
    arg_bridge_public_holiday = tr("Feriado con fines turísticos")

    # Bicentenary of the creation and first oath of the national flag.
    bicentennial_national_flag = tr(
        "Bicentenario de la creación y primera jura de la bandera nacional"
    )

    # Bicentenary of the Battle of Tucuman.
    bicentennial_battle_tucuman = tr("Bicentenario de la Batalla de Tucumán")

    # Bicentenary of the inaugural session of the National Constituent Assembly of the year 1813.
    bicentennial_assembly_1813 = tr(
        "Bicentenario de la sesión inaugural de la Asamblea Nacional Constituyente del año 1813"
    )

    # Bicentenary of the Battle of Salta.
    bicentennial_battle_salta = tr("Bicentenario de la Batalla de Salta")

    # National Census Day 2022.
    national_census_2022 = tr("Censo nacional 2022")

    special_holidays = {
        2011: (
            (MAR, 25, arg_bridge_public_holiday),
            (DEC, 9, arg_bridge_public_holiday),
        ),
        2012: (
            (FEB, 27, bicentennial_national_flag),
            (APR, 30, arg_bridge_public_holiday),
            (SEP, 24, bicentennial_battle_tucuman),
            (DEC, 24, arg_bridge_public_holiday),
        ),
        2013: (
            (JAN, 31, bicentennial_assembly_1813),
            (FEB, 20, bicentennial_battle_salta),
            (APR, 1, arg_bridge_public_holiday),
            (JUN, 21, arg_bridge_public_holiday),
        ),
        2014: (
            (MAY, 2, arg_bridge_public_holiday),
            (DEC, 26, arg_bridge_public_holiday),
        ),
        2015: (
            (MAR, 23, arg_bridge_public_holiday),
            (DEC, 7, arg_bridge_public_holiday),
        ),
        2016: (
            (JUL, 8, arg_bridge_public_holiday),
            (DEC, 9, arg_bridge_public_holiday),
        ),
        2018: (
            (APR, 30, arg_bridge_public_holiday),
            (DEC, 24, arg_bridge_public_holiday),
            (DEC, 31, arg_bridge_public_holiday),
        ),
        2019: (
            (JUL, 8, arg_bridge_public_holiday),
            (AUG, 19, arg_bridge_public_holiday),
            (OCT, 14, arg_bridge_public_holiday),
        ),
        2020: (
            (MAR, 23, arg_bridge_public_holiday),
            (JUL, 10, arg_bridge_public_holiday),
            (DEC, 7, arg_bridge_public_holiday),
        ),
        2021: (
            (MAY, 24, arg_bridge_public_holiday),
            (OCT, 8, arg_bridge_public_holiday),
            (NOV, 22, arg_bridge_public_holiday),
        ),
        2022: (
            (MAY, 18, national_census_2022),
            (OCT, 7, arg_bridge_public_holiday),
            (NOV, 21, arg_bridge_public_holiday),
            (DEC, 9, arg_bridge_public_holiday),
        ),
        2023: (
            (MAY, 26, arg_bridge_public_holiday),
            (JUN, 19, arg_bridge_public_holiday),
            (OCT, 13, arg_bridge_public_holiday),
        ),
    }

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _move_holiday(self, dt: date) -> None:
        """
         Movable Holidays Laws:
        - Decreto 1584/2010: 2010-11-03
            - AUG 17, OCT 12, NOV 20 Holidays will always be on MON
        - Decreto 52/2017: 2017-01-23 (Reconfirmed in Ley 27399)
            - If TUE/WED - observed on previous MON
            - If THU/FRI - observed on next MON
        """
        if self.observed:
            obs_date = None
            if self._is_tuesday(dt) or self._is_wednesday(dt):
                obs_date = self._get_nth_weekday_from(-1, MON, dt)
            elif self._is_thursday(dt) or self._is_friday(dt):
                obs_date = self._get_nth_weekday_from(1, MON, dt)
            if obs_date:
                self._add_holiday(self.tr("%s (Observado)") % self[dt], obs_date)
                self.pop(dt)

    def _populate(self, year):
        super()._populate(year)

        # Fixed Holidays

        # Status: In-Use.

        # New Year's Day.
        self._add_new_years_day(tr("Año Nuevo"))

        # Status: In-Use.
        # Started in 1956, abandoned in 1976.
        # Restarted in 2011 via Decreto 1584/2010.

        if 1956 <= year <= 1975 or year >= 2011:
            # Carnival.
            name = tr("Día de Carnaval")
            self._add_carnival_monday(name)
            self._add_carnival_tuesday(name)

        # Status: In-Use
        # Started in 2006, nearly reclassified as Movable Holidays in 2017

        if year >= 2006:
            self._add_holiday(
                # Memory's National Day for the Truth and Justice.
                tr("Día Nacional de la Memoria por la Verdad y la Justicia"),
                MAR,
                24,
            )

        # Status: In-Use.
        # Started in 1993 as War Veterans Day via Ley 24160.
        # Merged in 2001, confirmed as Fixed Holiday in 2006.
        # Superceded "Day of Argentine Sovereignty over the Malvinas".
        # Got moved temporary in 2020 (Decreto 297/2020).

        if year >= 1993:
            self._add_holiday(
                # War Veterans Day.
                tr("Día del Veterano de Guerra") if year <= 2000
                # Veterans Day and the Fallen in the Malvinas War.
                else tr("Día del Veterano y de los Caidos en la Guerra de Malvinas"),
                *((MAR, 31) if year == 2020 else (APR, 2)),
            )

        # Good Friday.
        self._add_good_friday(tr("Viernes Santo"))

        if year >= 1930:
            # Labor Day.
            self._add_labor_day(tr("Día del Trabajo"))

        if year >= 1813:
            # May Revolution Day.
            self._add_holiday(tr("Día de la Revolución de Mayo"), MAY, 25)

        # Status: Defunct.
        # Started in 1983 on April 2, moved to June 10 in Decreto 901/1984.
        # Abandoned in 2001.
        # Superceded by "Veterans Day and the Fallen in the Malvinas War".

        if 1983 <= year <= 2000:
            self._add_holiday(
                # Day of Argentine Sovereignty over the Malvinas, Sandwich and
                # South Atlantic Islands.
                tr(
                    "Día de los Derechos Argentinos sobre las Islas Malvinas, "
                    "Sandwich y del Atlántico Sur"
                ),
                *((APR, 2) if year == 1983 else (JUN, 10)),
            )

        # Also called "National Flag Day" (Día de la Bandera Nacional).
        # Status: In-Use.
        # Started in 1938 via Ley 12361 as Fixed Holiday.
        # Set as 3rd MON of JUN via Ley 24455 in Dec 1994.
        # Made Fixed Holiday again in 2011.

        if year >= 1938:
            self._add_holiday(
                # Pass to the Immortality of General Don Manuel Belgrano.
                tr("Paso a la Inmortalidad del General Don Manuel Belgrano"),
                self._get_nth_weekday_of_month(3, MON, JUN)
                if 1995 <= year <= 2010
                else date(year, JUN, 20),
            )

        if year >= 1816:
            # Independence Day.
            self._add_holiday(tr("Día de la Independencia"), JUL, 9)

        # Immaculate Conception.
        self._add_immaculate_conception_day(tr("Inmaculada Concepción de María"))

        # Christmas.
        self._add_christmas_day(tr("Navidad"))

        # Movable Holidays

        # Status: In-Use.
        # Started in 2014 for Salta, 2016 for the whole country via Ley 27258.

        if year >= 2016:
            dt = self._add_holiday(
                # Pass to the Immortality of General Don Martin Miguel de Guemes.
                tr("Paso a la Inmortalidad del General Don Martín Miguel de Güemes"),
                JUN,
                17,
            )
            # If Jun 17 is Friday, then it should move to Mon, Jun 20
            # but Jun 20 is Gen. Belgrano holiday
            if not self._is_friday(dt):
                self._move_holiday(dt)

        # Status: In-Use.
        # Started in 1938 via Ley 12387 on Aug 17.
        # Set as 3rd MON of AUG via Ley 24455 in Dec 1994.
        # Moved to Aug 22 for 2011 (election interfere) via Decreto 521/2011.

        # Pass to the Immortality of General Don Jose de San Martin.
        name = tr("Paso a la Inmortalidad del General Don José de San Martin")
        if year == 2011:
            self._add_holiday(name, AUG, 22)
        elif 1938 <= year <= 1994:
            self._add_holiday(name, AUG, 17)
        elif 1995 <= year <= 2010:
            self._add_holiday(name, self._get_nth_weekday_of_month(3, MON, AUG))
        elif year >= 2012:
            self._move_holiday(self._add_holiday(name, AUG, 17))

        # Status: In-Use.
        # First started in 1917 for Argentina.
        # In 2010 the holiday became movable and its name was changed.

        if year >= 1917:
            name = (
                # Respect for Cultural Diversity Day.
                tr("Día del Respeto a la Diversidad Cultural")
                if year >= 2010
                # Columbus Day.
                else tr("Día de la Raza")
            )
            dt = self._add_columbus_day(name)
            if year >= 2010:
                self._move_holiday(dt)

        # Status: In-Use.
        # First observed with no holiday via Ley 20770 in 1974.
        # Started in 2010.
        # Moved to Nov 27 for 2015 (election interfere).
        # Moved to Nov 28 again for 2016.

        if year >= 2010:
            # National Sovereignty Day.
            name = tr("Día de la Soberanía Nacional")
            if year == 2015:
                self._add_holiday(name, NOV, 27)
            elif year == 2016:
                self._add_holiday(name, NOV, 28)
            else:
                self._move_holiday(self._add_holiday(name, NOV, 20))


class AR(Argentina):
    pass


class ARG(Argentina):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td

from holidays.calendars.gregorian import FEB, APR, MAY, JUN, SEP, OCT
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Mozambique(HolidayBase, ChristianHolidays, InternationalHolidays):
    country = "MZ"

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_observed(self, dt: date) -> None:
        # whenever a public holiday falls on a Sunday,
        # it rolls over to the following Monday
        if self.observed and self._is_sunday(dt):
            self._add_holiday("%s (PONTE)" % self[dt], dt + td(days=+1))

    def _populate(self, year):
        if year <= 1974:
            return None

        super()._populate(year)

        self._add_observed(self._add_new_years_day("Ano novo"))

        self._add_good_friday("Sexta-feira Santa")

        self._add_carnival_tuesday("Carnaval")

        self._add_observed(self._add_holiday("Dia dos Heróis Moçambicanos", FEB, 3))

        self._add_observed(self._add_holiday("Dia da Mulher Moçambicana", APR, 7))

        self._add_observed(self._add_holiday("Dia Mundial do Trabalho", MAY, 1))

        self._add_observed(self._add_holiday("Dia da Independência Nacional", JUN, 25))

        self._add_observed(self._add_holiday("Dia da Vitória", SEP, 7))

        self._add_observed(self._add_holiday("Dia das Forças Armadas", SEP, 25))

        if year >= 1993:
            self._add_observed(self._add_holiday("Dia da Paz e Reconciliação", OCT, 4))

        self._add_observed(self._add_christmas_day("Dia de Natal e da Família"))


class MZ(Mozambique):
    pass


class MOZ(Mozambique):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

# flake8: noqa: F401

from holidays.calendars.gregorian import (
    JAN,
    FEB,
    MAR,
    APR,
    MAY,
    JUN,
    JUL,
    AUG,
    SEP,
    OCT,
    NOV,
    DEC,
    MON,
    TUE,
    WED,
    THU,
    FRI,
    SAT,
    SUN,
    WEEKEND,
)

HOLIDAY_NAME_DELIMITER = "; "  # Holiday names separator.

# Supported holiday categories.
BANK = "bank"
EXTENDED = "extended"
GOVERNMENT = "government"
HALF_DAY = "half_day"
PUBLIC = "public"
SCHOOL = "school"
WORKDAY = "workday"

CHINESE = "chinese"
CHRISTIAN = "christian"
HEBREW = "hebrew"
HINDU = "hindu"
ISLAMIC = "islamic"

ALL_CATEGORIES = {
    BANK,
    CHINESE,
    CHRISTIAN,
    EXTENDED,
    GOVERNMENT,
    HALF_DAY,
    HEBREW,
    HINDU,
    ISLAMIC,
    PUBLIC,
    SCHOOL,
    WORKDAY,
}



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.calendars import _CustomIslamicCalendar
from holidays.calendars.gregorian import APR, MAY, JUN, JUL, AUG, SEP, NOV, DEC, FRI, SAT
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import InternationalHolidays, IslamicHolidays


class UnitedArabEmirates(HolidayBase, InternationalHolidays, IslamicHolidays):
    """
    Holidays are regulated by the Article 74
    of Federal Law No. 08 for the year 1980:
    https://www.ilo.org/dyn/natlex/docs/ELECTRONIC/11956/69376/F417089305/ARE11956.pdf
    However the law is not applied literally,
    and was amended often in the past few years.
    Sources:
    2017: https://www.khaleejtimes.com/nation/uae-official-public-holidays-list-2017   # noqa: E501
    2018: https://www.thenational.ae/uae/government/uae-public-holidays-2018-announced-by-abu-dhabi-government-1.691393  # noqa: E501
    2019: https://www.thenational.ae/uae/government/uae-public-holidays-for-2019-and-2020-announced-by-cabinet-1.833425  # noqa: E501
    2020: https://u.ae/en/information-and-services/public-holidays-and-religious-affairs/public-holidays  # noqa: E501

    Holidays based on the Islamic Calendar are estimated (and so denoted),
    as are announced each year and based on moon sightings:
    - Eid al-Fitr
    - Eid al-Adha
    - Arafat (Hajj) Day
    - Al-Hijra (Islamic New Year
    - Mawlud al-Nabi (Prophet Mohammad's Birthday)
    - Leilat al-Miraj (Ascension of the Prophet), suspended after 2018.
    """

    country = "AE"
    default_language = "ar"
    # Estimated label.
    estimated_label = tr("(تقدير*) *%s")
    supported_languages = ("ar", "en_US")
    weekend = {FRI, SAT}

    def __init__(self, *args, **kwargs):
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self, calendar=UnitedArabEmiratesIslamicCalendar())
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("رأس السنة الميلادية"))

        if year >= 2015:
            # Commemoration Day.
            self._add_holiday(tr("يوم الشهيد"), *((DEC, 1) if year >= 2019 else (NOV, 30)))

        # National Day.
        self._add_holiday(tr("اليوم الوطني"), DEC, 2)
        self._add_holiday(tr("اليوم الوطني"), DEC, 3)

        # Eid al-Fitr.
        self._add_eid_al_fitr_day(tr("عيد الفطر"))
        # Eid al-Fitr Holiday.
        self._add_eid_al_fitr_day_two(tr("عطلة عيد الفطر"))
        self._add_eid_al_fitr_day_three(tr("عطلة عيد الفطر"))

        # Arafat Day.
        self._add_arafah_day(tr("وقفة عرفة"))

        # Eid al-Adha.
        self._add_eid_al_adha_day(tr("عيد الأضحى"))
        # Eid al-Adha Holiday.
        self._add_eid_al_adha_day_two(tr("عطلة عيد الأضحى"))
        self._add_eid_al_adha_day_three(tr("عطلة عيد الأضحى"))

        # Islamic New Year.
        self._add_islamic_new_year_day(tr("رأس السنة الهجرية"))

        # Leilat al-Miraj.
        if year <= 2018:  # The UAE government removed this starting 2019.
            self._add_isra_and_miraj_day(tr("ليلة المعراج"))

        # Prophet Muhammad's Birthday.
        if year <= 2019:  # The UAE government removed this starting 2020.
            self._add_mawlid_day(tr("عيد المولد النبوي"))


class AE(UnitedArabEmirates):
    pass


class ARE(UnitedArabEmirates):
    pass


class UnitedArabEmiratesIslamicCalendar(_CustomIslamicCalendar):
    EID_AL_ADHA_DATES = {
        2017: (SEP, 1),
        2018: (AUG, 21),
        2019: (AUG, 11),
        2020: (JUL, 31),
    }

    EID_AL_FITR_DATES = {
        2017: (JUN, 25),
        2018: (JUN, 14),
        2019: (JUN, 3),
        2020: (MAY, 24),
    }

    HIJRI_NEW_YEAR_DATES = {
        2017: (SEP, 22),
        2018: (SEP, 11),
        2019: (AUG, 31),
        2020: (AUG, 23),
    }

    ISRA_AND_MIRAJ_DATES = {
        2017: (APR, 23),
        2018: (APR, 13),
    }

    MAWLID_DATES = {
        2017: (NOV, 30),
        2018: (NOV, 19),
        2019: (NOV, 9),
    }



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars import _CustomIslamicCalendar
from holidays.calendars.gregorian import JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, NOV, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import (
    ChineseCalendarHolidays,
    ChristianHolidays,
    InternationalHolidays,
    IslamicHolidays,
)


class Brunei(
    HolidayBase, ChineseCalendarHolidays, ChristianHolidays, InternationalHolidays, IslamicHolidays
):
    """
    A subclass of :py:class:`HolidayBase` representing public holidays in Brunei Darussalam.

    References:

    - Based on: http://www.labour.gov.bn/Lists/Upcomming%20events/AllItems.aspx
                http://www.labour.gov.bn/Download/GUIDE%20TO%20BRUNEI%20EMPLOYMENT%20LAWS%20-%20english%20version-3.pdf  # noqa: E501
    - Checked with: https://asean.org/wp-content/uploads/2021/12/ASEAN-National-Holidays-2022.pdf
                    https://asean.org/wp-content/uploads/2022/12/ASEAN-Public-Holidays-2023.pdf
                    https://www.timeanddate.com/holidays/brunei/
    - [Jubli Emas Sultan Hassanal Bolkiah]
        https://www.brudirect.com/news.php?id=28316

    Limitations:

    - Brunei Darussalam holidays only works from 1984 onwards
    - Islamic holidays


    Country created by: `PPsyrius <https://github.com/PPsyrius>`__

    Country maintained by: `PPsyrius <https://github.com/PPsyrius>`__
    """

    country = "BN"
    default_language = "ms"
    estimated_label = tr("%s* (*anggaran)")
    supported_languages = ("en_US", "ms", "th")

    special_holidays = {
        # Sultan Hassanal Bolkiah's Golden Jubilee celebration
        2017: (OCT, 5, tr("Jubli Emas Sultan Hassanal Bolkiah")),
    }

    def __init__(self, *args, **kwargs):
        ChineseCalendarHolidays.__init__(self)
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self, calendar=BruneiIslamicCalendar())

        super().__init__(*args, **kwargs)

    def _populate(self, year):
        # Available post-Independence from 1984 afterwards
        if year <= 1983:
            return None

        def _add_observed(dt: date) -> None:
            """
            If Public Holiday falls on either Friday or Sunday, in-lieu
            observance is given out on the following Saturday or Monday.
            """
            if self.observed and (self._is_friday(dt) or self._is_sunday(dt)):
                for name in self.get_list(dt):
                    self._add_holiday(self.tr("%s - Diperhatikan") % name, dt + td(days=+1))

        super()._populate(year)

        # Awal Tahun Masihi
        # Status: In-Use.

        # New Year's Day
        _add_observed(self._add_new_years_day(tr("Awal Tahun Masihi")))

        # Tahun Baru Cina
        # Status: In-Use.

        # Lunar New Year
        _add_observed(self._add_chinese_new_years_day(tr("Tahun Baru Cina")))

        # Hari Kebangsaan
        # Status: In-Use.
        # Starts in 1984.

        # National Day
        _add_observed(self._add_holiday(tr("Hari Kebangsaan"), FEB, 23))

        # Hari Angkatan Bersenjata Diraja Brunei
        # Status: In-Use.
        # Starts in 1984.
        # Commemorates the formation of Royal Brunei Malay Regiment in 1961.

        _add_observed(
            # Armed Forces Day
            self._add_holiday(tr("Hari Angkatan Bersenjata Diraja Brunei"), MAY, 31)
        )

        # Hari Keputeraan KDYMM Sultan Brunei
        # Status: In-Use.
        # Started in 1968.

        _add_observed(
            # Sultan Hassanal Bolkiah's Birthday
            self._add_holiday(tr("Hari Keputeraan KDYMM Sultan Brunei"), JUL, 15)
        )

        # Hari Natal
        # Status: In-Use.

        # Christmas Day
        _add_observed(self._add_christmas_day(tr("Hari Natal")))

        # Islamic Holidays are placed after Gregorian holidays to prevent
        # the duplication of observed tags. - see #1168

        # Israk dan Mikraj
        # Status: In-Use.

        # Isra Mi'raj
        for dt in self._add_isra_and_miraj_day(tr("Israk dan Mikraj")):
            _add_observed(dt)

        # Hari Pertama Berpuasa
        # Status: In-Use.

        # First Day of Ramadan
        for dt in self._add_ramadan_beginning_day(tr("Hari Pertama Berpuasa")):
            _add_observed(dt)

        # Hari Nuzul Al-Quran
        # Status: In-Use.

        # Anniversary of the revelation of the Quran
        for dt in self._add_nuzul_al_quran_day(tr("Hari Nuzul Al-Quran")):
            _add_observed(dt)

        # Hari Raya Aidil Fitri
        # Status: In-Use.
        # This is celebrate for three days in Brunei.
        # Observed as 'Hari Raya Puasa' and only for 2 days at certain point.
        # We utilizes a separate in-lieu trigger for this one.
        # 1: If Wed-Thu-Fri -> Sat (+3)
        # 2: If Thu-Fri-Sat -> Mon (+4)
        # 3: If Fri-Sat-Sun -> Mon (+3)
        # 4: If Sat-Sun-Mon -> Tue (+3)
        # 5: If Sun-Mon-Tue -> Wed (+3)

        # Eid al-Fitr
        name = tr("Hari Raya Aidil Fitri")

        al_fitr_dates = self._add_eid_al_fitr_day(name)
        self._add_eid_al_fitr_day_two(name)
        self._add_eid_al_fitr_day_three(name)

        if self.observed:
            for dt in al_fitr_dates:
                obs_date = None
                for delta in range(3):
                    hol_date = dt + td(days=delta)
                    if self._is_friday(hol_date) or self._is_sunday(hol_date):
                        obs_date = dt + td(days=+3)
                        if self._is_sunday(obs_date):
                            obs_date += td(days=+1)
                        break
                if obs_date:
                    self._add_islamic_calendar_holiday(
                        self.tr("%s - Diperhatikan") % self[hol_date],
                        ((obs_date, self._year not in BruneiIslamicCalendar.EID_AL_FITR_DATES),),
                    )

        # Hari Raya Aidil Adha
        # Status: In-Use.

        # Eid al-Adha
        for dt in self._add_eid_al_adha_day(tr("Hari Raya Aidil Adha")):
            _add_observed(dt)

        # Awal Tahun Hijrah
        # Status: In-Use.

        # Islamic New Year
        for dt in self._add_islamic_new_year_day(tr("Awal Tahun Hijrah")):
            _add_observed(dt)

        # Maulidur Rasul
        # Status: In-Use.

        # Birth of the Prophet
        for dt in self._add_mawlid_day(tr("Maulidur Rasul")):
            _add_observed(dt)


class BN(Brunei):
    pass


class BRN(Brunei):
    pass


class BruneiIslamicCalendar(_CustomIslamicCalendar):
    EID_AL_ADHA_DATES = {
        2000: (MAR, 16),
        2001: (MAR, 6),
        2002: (FEB, 23),
        2003: (FEB, 12),
        2004: (FEB, 2),
        2005: (JAN, 21),
        2006: ((JAN, 10), (DEC, 31)),
        2007: (DEC, 20),
        2008: (DEC, 9),
        2009: (NOV, 28),
        2010: (NOV, 17),
        2011: (NOV, 7),
        2012: (OCT, 26),
        2013: (OCT, 15),
        2014: (OCT, 5),
        2015: (SEP, 24),
        2016: (SEP, 13),
        2017: (SEP, 2),
        2018: (AUG, 22),
        2019: (AUG, 11),
        2020: (AUG, 1),
        2021: (JUL, 20),
        2022: (JUL, 10),
        2023: (JUN, 29),
    }

    EID_AL_FITR_DATES = {
        2000: ((JAN, 8), (DEC, 28)),
        2001: (DEC, 17),
        2002: (DEC, 6),
        2003: (NOV, 26),
        2004: (NOV, 14),
        2005: (NOV, 3),
        2006: (OCT, 24),
        2007: (OCT, 13),
        2008: (OCT, 1),
        2009: (SEP, 20),
        2010: (SEP, 10),
        2011: (AUG, 31),
        2012: (AUG, 19),
        2013: (AUG, 8),
        2014: (JUL, 29),
        2015: (JUL, 18),
        2016: (JUL, 7),
        2017: (JUN, 26),
        2018: (JUN, 15),
        2019: (JUN, 5),
        2020: (MAY, 24),
        2021: (MAY, 13),
        2022: (MAY, 3),
        2023: (APR, 22),
    }

    HIJRI_NEW_YEAR_DATES = {
        2000: (APR, 6),
        2001: (MAR, 26),
        2002: (MAR, 15),
        2003: (MAR, 5),
        2004: (FEB, 22),
        2005: (FEB, 10),
        2006: (JAN, 31),
        2007: (JAN, 20),
        2008: ((JAN, 10), (DEC, 29)),
        2009: (DEC, 18),
        2010: (DEC, 8),
        2011: (NOV, 27),
        2012: (NOV, 15),
        2013: (NOV, 5),
        2014: (OCT, 25),
        2015: (OCT, 15),
        2016: (OCT, 3),
        2017: (SEP, 22),
        2018: (SEP, 12),
        2019: (SEP, 1),
        2020: (AUG, 20),
        2021: (AUG, 10),
        2022: (JUL, 30),
        2023: (JUL, 19),
    }

    ISRA_AND_MIRAJ_DATES = {
        2000: (OCT, 26),
        2001: (OCT, 15),
        2002: (OCT, 4),
        2003: (SEP, 24),
        2004: (SEP, 12),
        2005: (SEP, 1),
        2006: (AUG, 22),
        2007: (AUG, 11),
        2008: (JUL, 31),
        2009: (JUL, 20),
        2010: (JUL, 9),
        2011: (JUN, 29),
        2012: (JUN, 17),
        2013: (JUN, 6),
        2014: (MAY, 27),
        2015: (MAY, 16),
        2016: (MAY, 5),
        2017: (APR, 24),
        2018: (APR, 14),
        2019: (APR, 3),
        2020: (MAR, 22),
        2021: (MAR, 11),
        2022: (MAR, 1),
        2023: (FEB, 18),
    }

    MAWLID_DATES = {
        2000: (JUN, 15),
        2001: (JUN, 4),
        2002: (MAY, 24),
        2003: (MAY, 14),
        2004: (MAY, 2),
        2005: (APR, 21),
        2006: (APR, 11),
        2007: (MAR, 31),
        2008: (MAR, 20),
        2009: (MAR, 9),
        2010: (FEB, 26),
        2011: (FEB, 16),
        2012: (FEB, 5),
        2013: (JAN, 24),
        2014: (JAN, 14),
        2015: ((JAN, 3), (DEC, 24)),
        2016: (DEC, 12),
        2017: (DEC, 1),
        2018: (NOV, 21),
        2019: (NOV, 9),
        2020: (OCT, 29),
        2021: (OCT, 19),
        2022: (OCT, 8),
        2023: (SEP, 28),
    }

    NUZUL_AL_QURAN_DATES = {
        2000: (DEC, 14),
        2001: (DEC, 3),
        2002: (NOV, 22),
        2003: (NOV, 12),
        2004: (NOV, 1),
        2005: (OCT, 21),
        2006: (OCT, 10),
        2007: (SEP, 29),
        2008: (SEP, 18),
        2009: (SEP, 7),
        2010: (AUG, 27),
        2011: (AUG, 17),
        2012: (AUG, 5),
        2013: (JUL, 25),
        2014: (JUL, 15),
        2015: (JUL, 4),
        2016: (JUN, 23),
        2017: (JUN, 12),
        2018: (JUN, 2),
        2019: (MAY, 23),
        2020: (MAY, 10),
        2021: (APR, 29),
        2022: (APR, 19),
        2023: (APR, 8),
    }

    RAMADAN_BEGINNING_DATES = {
        2000: (NOV, 28),
        2001: (NOV, 17),
        2002: (NOV, 6),
        2003: (OCT, 27),
        2004: (OCT, 16),
        2005: (OCT, 5),
        2006: (SEP, 24),
        2007: (SEP, 13),
        2008: (SEP, 2),
        2009: (AUG, 22),
        2010: (AUG, 11),
        2011: (AUG, 1),
        2012: (JUL, 20),
        2013: (JUL, 9),
        2014: (JUN, 29),
        2015: (JUN, 18),
        2016: (JUN, 7),
        2017: (MAY, 27),
        2018: (MAY, 16),
        2019: (MAY, 6),
        2020: (APR, 25),
        2021: (APR, 13),
        2022: (APR, 3),
        2023: (MAR, 23),
    }



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from holidays.calendars.gregorian import MAR, APR, MAY, JUL, OCT
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Lesotho(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    References:
    - https://en.wikipedia.org/wiki/Public_holidays_in_Lesotho
    - https://www.ilo.org/dyn/travail/docs/2093/Public%20Holidays%20Act%201995.pdf
    - https://www.timeanddate.com/holidays/lesotho/
    """

    country = "LS"
    special_holidays = {
        2002: (MAY, 25, "Africa Day"),
    }

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        if year <= 1995:
            return None

        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day("New Year's Day")

        # Moshoeshoe's Day.
        self._add_holiday("Moshoeshoe's Day", MAR, 11)

        if year <= 2002:
            # Heroes Day.
            self._add_holiday("Heroes Day", APR, 4)

        if year >= 2003:
            # Africa/Heroes Day.
            self._add_africa_day("Africa/Heroes Day")

        # Good Friday.
        self._add_good_friday("Good Friday")

        # Easter Monday.
        self._add_easter_monday("Easter Monday")

        # Workers' Day.
        self._add_labor_day("Workers' Day")

        # Ascension Day.
        self._add_ascension_thursday("Ascension Day")

        # https://en.wikipedia.org/wiki/Letsie_III
        # King's Birthday.
        self._add_holiday("King's Birthday", *((JUL, 17) if year >= 1998 else (MAY, 2)))

        # Independence Day.
        self._add_holiday("Independence Day", OCT, 4)

        # Christmas Day.
        self._add_christmas_day("Christmas Day")

        # Boxing Day.
        self._add_christmas_day_two("Boxing Day")


class LS(Lesotho):
    pass


class LSO(Lesotho):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from typing import Tuple, Union

from holidays.calendars.gregorian import (
    JAN,
    FEB,
    MAR,
    APR,
    MAY,
    JUN,
    JUL,
    AUG,
    SEP,
    OCT,
    NOV,
    DEC,
    MON,
    TUE,
    THU,
    FRI,
)
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class UnitedStates(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_the_United_States

    For Northern Mariana Islands (subdivision MP):
    https://governor.gov.mp/archived-news/executive-actions-archive/memorandum-2022-legal-holidays/
    https://webcache.googleusercontent.com/search?q=cache:C17_7FBgPtQJ:https://governor.gov.mp/archived-news/executive-actions-archive/memorandum-2022-legal-holidays/&hl=en&gl=sg&strip=1&vwsrc=0
    """

    country = "US"
    subdivisions: Union[Tuple[()], Tuple[str, ...]] = (
        "AK",
        "AL",
        "AR",
        "AS",
        "AZ",
        "CA",
        "CO",
        "CT",
        "DC",
        "DE",
        "FL",
        "FM",
        "GA",
        "GU",
        "HI",
        "IA",
        "ID",
        "IL",
        "IN",
        "KS",
        "KY",
        "LA",
        "MA",
        "MD",
        "ME",
        "MH",
        "MI",
        "MN",
        "MO",
        "MP",
        "MS",
        "MT",
        "NC",
        "ND",
        "NE",
        "NH",
        "NJ",
        "NM",
        "NV",
        "NY",
        "OH",
        "OK",
        "OR",
        "PA",
        "PR",
        "PW",
        "RI",
        "SC",
        "SD",
        "TN",
        "TX",
        "UM",
        "UT",
        "VA",
        "VI",
        "VT",
        "WA",
        "WI",
        "WV",
        "WY",
    )

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_observed(self, dt: date, before: bool = True, after: bool = True) -> None:
        if not self.observed:
            return None
        if self._is_saturday(dt) and before:
            self._add_holiday("%s (Observed)" % self[dt], dt + td(days=-1))
        elif self._is_sunday(dt) and after:
            self._add_holiday("%s (Observed)" % self[dt], dt + td(days=+1))

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day
        if year >= 1871:
            name = "New Year's Day"
            self._add_observed(self._add_new_years_day(name), before=False)
            # The following year's observed New Year's Day can be in this year
            # when it falls on a Friday (Jan 1st is a Saturday).
            if self.observed and self._is_friday(DEC, 31):
                self._add_holiday("%s (Observed)" % name, DEC, 31)

        # Memorial Day
        if year >= 1888:
            self._add_holiday(
                "Memorial Day",
                self._get_nth_weekday_of_month(-1, MON, MAY)
                if year >= 1971
                else date(year, MAY, 30),
            )

        # Juneteenth Day
        if year >= 2021:
            self._add_observed(self._add_holiday("Juneteenth National Independence Day", JUN, 19))

        # Independence Day
        if year >= 1871:
            self._add_observed(self._add_holiday("Independence Day", JUL, 4))

        # Labor Day
        if year >= 1894:
            self._add_holiday("Labor Day", self._get_nth_weekday_of_month(1, MON, SEP))

        # Veterans Day
        if year >= 1938:
            name = "Veterans Day" if year >= 1954 else "Armistice Day"
            if 1971 <= year <= 1977:
                self._add_holiday(name, self._get_nth_weekday_of_month(4, MON, OCT))
            else:
                self._add_observed(self._add_remembrance_day(name))

        # Thanksgiving
        if year >= 1871:
            self._add_holiday("Thanksgiving", self._get_nth_weekday_of_month(4, THU, NOV))

        # Christmas Day
        if year >= 1871:
            self._add_observed(self._add_christmas_day("Christmas Day"))

    def _add_christmas_eve_holiday(self):
        # Christmas Eve
        name = "Christmas Eve"
        dt = self._add_christmas_eve(name)
        if self.observed:
            # If on Friday, observed on Thursday
            if self._is_friday(dt):
                self._add_holiday("%s (Observed)" % name, dt + td(days=-1))
            # If on Saturday or Sunday, observed on Friday
            elif self._is_weekend(dt):
                self._add_holiday("%s (Observed)" % name, self._get_nth_weekday_from(-1, FRI, dt))

    def _add_subdiv_holidays(self):
        # Martin Luther King Jr. Day
        if self._year >= 1986 and self.subdiv not in {"AL", "AR", "AZ", "GA", "ID", "MS", "NH"}:
            self._add_holiday(
                "Martin Luther King Jr. Day", self._get_nth_weekday_of_month(3, MON, JAN)
            )

        # Washington's Birthday
        if self._year >= 1879 and self.subdiv not in {
            "AL",
            "AR",
            "DE",
            "FL",
            "GA",
            "NM",
            "PR",
            "VI",
        }:
            self._add_holiday(
                "Washington's Birthday",
                self._get_nth_weekday_of_month(3, MON, FEB)
                if self._year >= 1971
                else date(self._year, FEB, 22),
            )

        # Columbus Day
        if self._year >= 1937 and self.subdiv not in {
            "AK",
            "AR",
            "DE",
            "FL",
            "HI",
            "NV",
            "SD",
            "VI",
        }:
            name = "Columbus Day"
            if self._year >= 1970:
                self._add_holiday(name, self._get_nth_weekday_of_month(2, MON, OCT))
            else:
                self._add_columbus_day(name)

        super()._add_subdiv_holidays()

    def _add_subdiv_ak_holidays(self):
        # Seward's Day
        if self._year >= 1918:
            self._add_holiday(
                "Seward's Day",
                self._get_nth_weekday_of_month(-1, MON, MAR)
                if self._year >= 1955
                else date(self._year, MAR, 30),
            )

        # Alaska Day
        if self._year >= 1867:
            self._add_observed(self._add_holiday("Alaska Day", OCT, 18))

    def _add_subdiv_al_holidays(self):
        # Martin Luther King Jr. Day
        if self._year >= 1986:
            self._add_holiday(
                "Martin Luther King, Jr & Robert E. Lee's Birthday",
                self._get_nth_weekday_of_month(3, MON, JAN),
            )

        # Washington's Birthday
        self._add_holiday(
            "George Washington & Thomas Jefferson's Birthday",
            self._get_nth_weekday_of_month(3, MON, FEB)
            if self._year >= 1971
            else date(self._year, FEB, 22),
        )

        # Confederate Memorial Day
        if self._year >= 1866:
            self._add_holiday(
                "Confederate Memorial Day", self._get_nth_weekday_of_month(4, MON, APR)
            )

        # Jefferson Davis Birthday
        if self._year >= 1890:
            self._add_holiday(
                "Jefferson Davis Birthday", self._get_nth_weekday_of_month(1, MON, JUN)
            )

    def _add_subdiv_ar_holidays(self):
        # Martin Luther King Jr. Day
        if self._year >= 1986:
            self._add_holiday(
                "Martin Luther King Jr. Day"
                if self._year >= 2018
                else "Dr. Martin Luther King Jr. " "and Robert E. Lee's Birthdays",
                self._get_nth_weekday_of_month(3, MON, JAN),
            )

        # Washington's Birthday
        self._add_holiday(
            "George Washington's Birthday and Daisy Gatson Bates Day",
            self._get_nth_weekday_of_month(3, MON, FEB)
            if self._year >= 1971
            else date(self._year, FEB, 22),
        )

    def _add_subdiv_as_holidays(self):
        # Christmas Eve
        self._add_christmas_eve_holiday()

    def _add_subdiv_az_holidays(self):
        # Martin Luther King Jr. Day
        if self._year >= 1986:
            self._add_holiday(
                "Dr. Martin Luther King Jr. / Civil Rights Day",
                self._get_nth_weekday_of_month(3, MON, JAN),
            )

    def _add_subdiv_ca_holidays(self):
        # Lincoln's Birthday
        if 1971 <= self._year <= 2009:
            self._add_observed(self._add_holiday("Lincoln's Birthday", FEB, 12))

        # Susan B. Anthony Day
        if self._year >= 2014:
            self._add_holiday("Susan B. Anthony Day", FEB, 15)

        # Cesar Chavez Day
        if self._year >= 1995:
            self._add_observed(self._add_holiday("Cesar Chavez Day", MAR, 31), before=False)

        # Day After Thanksgiving
        if self._year >= 1975:
            self._add_holiday(
                "Day After Thanksgiving",
                self._get_nth_weekday_of_month(4, THU, NOV) + td(days=+1),
            )

    def _add_subdiv_co_holidays(self):
        pass

    def _add_subdiv_ct_holidays(self):
        # Lincoln's Birthday
        if self._year >= 1971:
            self._add_observed(self._add_holiday("Lincoln's Birthday", FEB, 12))

        # Good Friday
        self._add_good_friday("Good Friday")

    def _add_subdiv_dc_holidays(self):
        # Inauguration Day
        if self._year >= 1789 and (self._year - 1789) % 4 == 0:
            dt = (JAN, 20) if self._year >= 1937 else (MAR, 4)
            self._add_observed(self._add_holiday("Inauguration Day", *dt), before=False)

        # Emancipation Day
        if self._year >= 2005:
            self._add_observed(self._add_holiday("Emancipation Day", APR, 16))

    def _add_subdiv_de_holidays(self):
        # Good Friday
        self._add_good_friday("Good Friday")

        # Election Day
        if self._year >= 2008 and self._year % 2 == 0:
            self._add_holiday(
                "Election Day",
                self._get_nth_weekday_of_month(1, MON, NOV) + td(days=+1),
            )

        # Day After Thanksgiving
        if self._year >= 1975:
            self._add_holiday(
                "Day After Thanksgiving",
                self._get_nth_weekday_of_month(4, THU, NOV) + td(days=+1),
            )

    def _add_subdiv_fl_holidays(self):
        # Susan B. Anthony Day
        if self._year >= 2011:
            self._add_holiday("Susan B. Anthony Day", FEB, 15)

        # Friday After Thanksgiving
        if self._year >= 1975:
            self._add_holiday(
                "Friday After Thanksgiving",
                self._get_nth_weekday_of_month(4, THU, NOV) + td(days=+1),
            )

    def _add_subdiv_fm_holidays(self):
        pass

    def _add_subdiv_ga_holidays(self):
        # Martin Luther King Jr. Day
        if self._year >= 1986:
            self._add_holiday(
                "Martin Luther King Jr. Day" if self._year >= 2012 else "Robert E. Lee's Birthday",
                self._get_nth_weekday_of_month(3, MON, JAN),
            )

        # Confederate Memorial Day
        if self._year >= 1866:
            self._add_holiday(
                "State Holiday" if self._year >= 2016 else "Confederate Memorial Day",
                date(self._year, APR, 10)
                if self._year == 2020
                else self._get_nth_weekday_of_month(4, MON, APR),
            )

        # Robert E. Lee's Birthday
        if self._year >= 1986:
            self._add_holiday(
                "State Holiday" if self._year >= 2016 else "Robert E. Lee's Birthday",
                self._get_nth_weekday_of_month(4, THU, NOV) + td(days=+1),
            )

        # Washington's Birthday
        dt = (DEC, 24)
        if self._is_wednesday(*dt):
            dt = (DEC, 26)
        self._add_holiday("Washington's Birthday", *dt)

    def _add_subdiv_gu_holidays(self):
        # Guam Discovery Day
        if self._year >= 1970:
            self._add_holiday("Guam Discovery Day", self._get_nth_weekday_of_month(1, MON, MAR))

        # Good Friday
        self._add_good_friday("Good Friday")

        # Liberation Day (Guam)
        if self._year >= 1945:
            self._add_holiday("Liberation Day (Guam)", JUL, 21)

        # All Souls' Day
        self._add_all_souls_day("All Souls' Day")

        # Lady of Camarin Day
        self._add_immaculate_conception_day("Lady of Camarin Day")

    def _add_subdiv_hi_holidays(self):
        # Prince Jonah Kuhio Kalanianaole Day
        if self._year >= 1949:
            self._add_observed(self._add_holiday("Prince Jonah Kuhio Kalanianaole Day", MAR, 26))

        # Kamehameha Day
        if self._year >= 1872:
            jun_11 = self._add_holiday("Kamehameha Day", JUN, 11)
            if self._year >= 2011:
                self._add_observed(jun_11)

        # Statehood Day
        if self._year >= 1959:
            self._add_holiday("Statehood Day", self._get_nth_weekday_of_month(3, FRI, AUG))

        # Election Day
        if self._year >= 2008 and self._year % 2 == 0:
            self._add_holiday(
                "Election Day",
                self._get_nth_weekday_of_month(1, MON, NOV) + td(days=+1),
            )

    def _add_subdiv_ia_holidays(self):
        # Lincoln's Birthday
        if self._year >= 1971:
            self._add_observed(self._add_holiday("Lincoln's Birthday", FEB, 12))

    def _add_subdiv_id_holidays(self):
        # Martin Luther King Jr. Day
        if self._year >= 1986:
            self._add_holiday(
                "Martin Luther King Jr. / Idaho Human Rights Day"
                if self._year >= 2006
                else "Martin Luther King Jr. Day",
                self._get_nth_weekday_of_month(3, MON, JAN),
            )

    def _add_subdiv_il_holidays(self):
        # Lincoln's Birthday
        if self._year >= 1971:
            self._add_observed(self._add_holiday("Lincoln's Birthday", FEB, 12))

        # Casimir Pulaski Day
        if self._year >= 1978:
            self._add_holiday("Casimir Pulaski Day", self._get_nth_weekday_of_month(1, MON, MAR))

        # Election Day
        if self._year >= 2008 and self._year % 2 == 0:
            self._add_holiday(
                "Election Day",
                self._get_nth_weekday_of_month(1, MON, NOV) + td(days=+1),
            )

    def _add_subdiv_in_holidays(self):
        # Good Friday
        self._add_good_friday("Good Friday")

        # Primary Election Day
        if self._year >= 2015 or (self._year >= 2006 and self._year % 2 == 0):
            self._add_holiday(
                "Primary Election Day",
                self._get_nth_weekday_of_month(1, MON, MAY) + td(days=+1),
            )

        # Election Day
        if self._year >= 2015 or (self._year >= 2008 and self._year % 2 == 0):
            self._add_holiday(
                "Election Day",
                self._get_nth_weekday_of_month(1, MON, NOV) + td(days=+1),
            )

        # Lincoln's Birthday
        if self._year >= 2010:
            self._add_holiday(
                "Lincoln's Birthday",
                self._get_nth_weekday_of_month(4, THU, NOV) + td(days=+1),
            )

    def _add_subdiv_ks_holidays(self):
        # Christmas Eve
        if self._year >= 2013:
            self._add_christmas_eve_holiday()

    def _add_subdiv_ky_holidays(self):
        # Good Friday
        self._add_good_friday("Good Friday")

        # New Year's Eve
        if self._year >= 2013:
            self._add_observed(self._add_new_years_eve("New Year's Eve"), after=False)

    def _add_subdiv_la_holidays(self):
        # Inauguration Day
        if self._year >= 1789 and (self._year - 1789) % 4 == 0:
            dt = (JAN, 20) if self._year >= 1937 else (MAR, 4)
            self._add_observed(self._add_holiday("Inauguration Day", *dt), before=False)

        # Mardi Gras
        if self._year >= 1857:
            self._add_carnival_tuesday("Mardi Gras")

        # Good Friday
        self._add_good_friday("Good Friday")

        # Election Day
        if self._year >= 2008 and self._year % 2 == 0:
            self._add_holiday(
                "Election Day",
                self._get_nth_weekday_of_month(1, MON, NOV) + td(days=+1),
            )

    def _add_subdiv_ma_holidays(self):
        # Evacuation Day
        if self._year >= 1901:
            mar_17 = self._add_holiday("Evacuation Day", MAR, 17)
            if self.observed and self._is_weekend(mar_17):
                self._add_holiday(
                    "%s (Observed)" % self[mar_17], self._get_nth_weekday_from(1, MON, mar_17)
                )

        # Patriots' Day
        if self._year >= 1894:
            self._add_holiday(
                "Patriots' Day",
                self._get_nth_weekday_of_month(3, MON, APR)
                if self._year >= 1969
                else date(self._year, APR, 19),
            )

    def _add_subdiv_md_holidays(self):
        # Inauguration Day
        if self._year >= 1789 and (self._year - 1789) % 4 == 0:
            dt = (JAN, 20) if self._year >= 1937 else (MAR, 4)
            self._add_observed(self._add_holiday("Inauguration Day", *dt), before=False)

        # American Indian Heritage Day
        if self._year >= 2008:
            self._add_holiday(
                "American Indian Heritage Day",
                self._get_nth_weekday_of_month(4, THU, NOV) + td(days=+1),
            )

    def _add_subdiv_me_holidays(self):
        # Patriots' Day
        if self._year >= 1894:
            self._add_holiday(
                "Patriots' Day",
                self._get_nth_weekday_of_month(3, MON, APR)
                if self._year >= 1969
                else date(self._year, APR, 19),
            )

    def _add_subdiv_mh_holidays(self):
        pass

    def _add_subdiv_mi_holidays(self):
        if self._year >= 2013:
            # Christmas Eve
            self._add_christmas_eve_holiday()

            # New Year's Eve
            self._add_observed(self._add_new_years_eve("New Year's Eve"), after=False)

    def _add_subdiv_mn_holidays(self):
        pass

    def _add_subdiv_mo_holidays(self):
        # Truman Day
        if self._year >= 1949:
            self._add_observed(self._add_holiday("Truman Day", MAY, 8))

    def _add_subdiv_mp_holidays(self):
        # Commonwealth Covenant Day in Northern Mariana Islands
        self._add_observed(self._add_holiday("Commonwealth Covenant Day", MAR, 24))

        # Good Friday
        self._add_good_friday("Good Friday")

        # Commonwealth Cultural Day in Northern Mariana Islands
        self._add_holiday("Commonwealth Cultural Day", self._get_nth_weekday_of_month(2, MON, OCT))

        # Election Day
        if self._year >= 2008 and self._year % 2 == 0:
            self._add_holiday(
                "Election Day",
                self._get_nth_weekday_of_month(1, MON, NOV) + td(days=+1),
            )

        # Citizenship Day in Northern Mariana Islands
        self._add_observed(self._add_holiday("Citizenship Day", NOV, 4))

        # Constitution Day in Northern Mariana Islands
        self._add_observed(self._add_holiday("Constitution Day", DEC, 8))

    def _add_subdiv_ms_holidays(self):
        # Martin Luther King Jr. Day
        if self._year >= 1986:
            self._add_holiday(
                "Dr. Martin Luther King Jr. and Robert E. Lee's Birthdays",
                self._get_nth_weekday_of_month(3, MON, JAN),
            )

        # Confederate Memorial Day
        if self._year >= 1866:
            self._add_holiday(
                "Confederate Memorial Day", self._get_nth_weekday_of_month(4, MON, APR)
            )

    def _add_subdiv_mt_holidays(self):
        # Election Day
        if self._year >= 2008 and self._year % 2 == 0:
            self._add_holiday(
                "Election Day",
                self._get_nth_weekday_of_month(1, MON, NOV) + td(days=+1),
            )

    def _add_subdiv_nc_holidays(self):
        # Good Friday
        self._add_good_friday("Good Friday")

        # Day After Thanksgiving
        if self._year >= 1975:
            self._add_holiday(
                "Day After Thanksgiving",
                self._get_nth_weekday_of_month(4, THU, NOV) + td(days=+1),
            )

        # Christmas Eve
        if self._year >= 2013:
            self._add_christmas_eve_holiday()

        # Day After Christmas
        if self._year >= 2013:
            name = "Day After Christmas"
            dt = self._add_christmas_day_two(name)
            if self.observed:
                # If on Saturday or Sunday, observed on Monday
                if self._is_weekend(dt):
                    self._add_holiday(
                        "%s (Observed)" % name, self._get_nth_weekday_from(1, MON, dt)
                    )
                # If on Monday, observed on Tuesday
                elif self._is_monday(dt):
                    self._add_holiday("%s (Observed)" % name, dt + td(days=+1))

    def _add_subdiv_nd_holidays(self):
        pass

    def _add_subdiv_ne_holidays(self):
        # Arbor Day
        if self._year >= 1875:
            self._add_holiday(
                "Arbor Day",
                self._get_nth_weekday_of_month(-1, FRI, APR)
                if self._year >= 1989
                else date(self._year, APR, 22),
            )

    def _add_subdiv_nh_holidays(self):
        # Martin Luther King Jr. Day
        if self._year >= 1986:
            self._add_holiday(
                "Dr. Martin Luther King Jr. / Civil Rights Day",
                self._get_nth_weekday_of_month(3, MON, JAN),
            )

        # Election Day
        if self._year >= 2008 and self._year % 2 == 0:
            self._add_holiday(
                "Election Day",
                self._get_nth_weekday_of_month(1, MON, NOV) + td(days=+1),
            )

        # Day After Thanksgiving
        if self._year >= 1975:
            self._add_holiday(
                "Day After Thanksgiving",
                self._get_nth_weekday_of_month(4, THU, NOV) + td(days=+1),
            )

    def _add_subdiv_nj_holidays(self):
        # Lincoln's Birthday
        if self._year >= 1971:
            self._add_observed(self._add_holiday("Lincoln's Birthday", FEB, 12))

        # Good Friday
        self._add_good_friday("Good Friday")

        # Election Day
        if self._year >= 2008 and self._year % 2 == 0:
            self._add_holiday(
                "Election Day",
                self._get_nth_weekday_of_month(1, MON, NOV) + td(days=+1),
            )

    def _add_subdiv_nm_holidays(self):
        # Presidents' Day
        self._add_holiday(
            "Presidents' Day",
            self._get_nth_weekday_of_month(4, THU, NOV) + td(days=+1),
        )

    def _add_subdiv_nv_holidays(self):
        # Nevada Day
        if self._year >= 1933:
            self._add_observed(
                self._add_holiday(
                    "Nevada Day",
                    self._get_nth_weekday_of_month(-1, FRI, OCT)
                    if self._year >= 2000
                    else date(self._year, OCT, 31),
                )
            )

        # Family Day
        self._add_holiday("Family Day", self._get_nth_weekday_of_month(4, THU, NOV) + td(days=+1))

    def _add_subdiv_ny_holidays(self):
        # Lincoln's Birthday
        if self._year >= 1971:
            self._add_observed(self._add_holiday("Lincoln's Birthday", FEB, 12))

        # Susan B. Anthony Day
        if self._year >= 2004:
            self._add_holiday("Susan B. Anthony Day", FEB, 15)

        # Election Day
        if self._year >= 2015 or (self._year >= 2008 and self._year % 2 == 0):
            self._add_holiday(
                "Election Day",
                self._get_nth_weekday_of_month(1, MON, NOV) + td(days=+1),
            )

    def _add_subdiv_oh_holidays(self):
        pass

    def _add_subdiv_ok_holidays(self):
        # Day After Thanksgiving
        if self._year >= 1975:
            self._add_holiday(
                "Day After Thanksgiving",
                self._get_nth_weekday_of_month(4, THU, NOV) + td(days=+1),
            )

    def _add_subdiv_or_holidays(self):
        pass

    def _add_subdiv_pa_holidays(self):
        # Day After Thanksgiving
        self._add_holiday(
            "Day After Thanksgiving",
            self._get_nth_weekday_of_month(4, THU, NOV) + td(days=+1),
        )

    def _add_subdiv_pr_holidays(self):
        # Epiphany
        self._add_epiphany_day("Epiphany")

        # Washington's Birthday
        self._add_holiday("Presidents' Day", self._get_nth_weekday_of_month(3, MON, FEB))

        # Emancipation Day
        self._add_observed(self._add_holiday("Emancipation Day", MAR, 22), before=False)

        # Good Friday
        self._add_good_friday("Good Friday")

        # Constitution Day
        self._add_observed(self._add_holiday("Constitution Day", JUL, 25), before=False)

        # Discovery Day
        self._add_observed(self._add_holiday("Discovery Day", NOV, 19), before=False)

    def _add_subdiv_pw_holidays(self):
        pass

    def _add_subdiv_ri_holidays(self):
        # Victory Day
        if self._year >= 1948:
            self._add_holiday("Victory Day", self._get_nth_weekday_of_month(2, MON, AUG))

    def _add_subdiv_sc_holidays(self):
        # Confederate Memorial Day
        if self._year >= 1866:
            self._add_holiday(
                "Confederate Memorial Day", self._get_nth_weekday_of_month(4, MON, APR)
            )

    def _add_subdiv_sd_holidays(self):
        # Columbus Day
        if self._year >= 1937:
            name = "Native American Day"
            if self._year >= 1970:
                self._add_holiday(name, self._get_nth_weekday_of_month(2, MON, OCT))
            else:
                self._add_columbus_day(name)

    def _add_subdiv_tn_holidays(self):
        # Good Friday
        self._add_good_friday("Good Friday")

    def _add_subdiv_tx_holidays(self):
        # Confederate Memorial Day
        if self._year >= 1931:
            self._add_holiday("Confederate Memorial Day", JAN, 19)

        # Texas Independence Day
        if self._year >= 1874:
            self._add_holiday("Texas Independence Day", MAR, 2)

        # Cesar Chavez Day
        if self._year >= 2000:
            self._add_holiday("Cesar Chavez Day", MAR, 31)

        # Good Friday
        self._add_good_friday("Good Friday")

        # San Jacinto Day
        if self._year >= 1875:
            self._add_holiday("San Jacinto Day", APR, 21)

        # Emancipation Day In Texas
        if self._year >= 1980:
            self._add_holiday("Emancipation Day In Texas", JUN, 19)

        # Lyndon Baines Johnson Day
        if self._year >= 1973:
            self._add_holiday("Lyndon Baines Johnson Day", AUG, 27)

        # Friday After Thanksgiving
        if self._year >= 1975:
            self._add_holiday(
                "Friday After Thanksgiving",
                self._get_nth_weekday_of_month(4, THU, NOV) + td(days=+1),
            )

        # Christmas Eve
        if self._year >= 1981:
            self._add_christmas_eve_holiday()

        # Day After Christmas
        if self._year >= 1981:
            self._add_christmas_day_two("Day After Christmas")

    def _add_subdiv_um_holidays(self):
        pass

    def _add_subdiv_ut_holidays(self):
        # Pioneer Day
        if self._year >= 1849:
            self._add_observed(self._add_holiday("Pioneer Day", JUL, 24))

    def _add_subdiv_va_holidays(self):
        # Lee Jackson Day
        if 1889 <= self._year <= 2020:
            if self._year >= 2000:
                dt = self._get_nth_weekday_of_month(3, MON, JAN) + td(days=-3)
            elif self._year >= 1983:
                dt = self._get_nth_weekday_of_month(3, MON, JAN)
            else:
                dt = date(self._year, JAN, 19)
            self._add_holiday("Lee Jackson Day", dt)

        # Inauguration Day
        if self._year >= 1789 and (self._year - 1789) % 4 == 0:
            dt = (JAN, 20) if self._year >= 1937 else (MAR, 4)
            self._add_observed(self._add_holiday("Inauguration Day", *dt), before=False)

    def _add_subdiv_vi_holidays(self):
        # Three Kings Day
        self._add_epiphany_day("Three Kings Day")

        # Washington's Birthday
        self._add_holiday(
            "Presidents' Day",
            self._get_nth_weekday_of_month(3, MON, FEB)
            if self._year >= 1971
            else date(self._year, FEB, 22),
        )

        # Transfer Day
        self._add_holiday("Transfer Day", MAR, 31)

        # Holy Thursday
        self._add_holy_thursday("Holy Thursday")

        # Good Friday
        self._add_good_friday("Good Friday")

        # Easter Monday
        self._add_easter_monday("Easter Monday")

        # Emancipation Day in US Virgin Islands
        self._add_holiday("Emancipation Day", JUL, 3)

        # Columbus Day
        if self._year >= 1937:
            name = "Columbus Day and Puerto Rico Friendship Day"
            if self._year >= 1970:
                self._add_holiday(name, self._get_nth_weekday_of_month(2, MON, OCT))
            else:
                self._add_columbus_day(name)

        # Liberty Day
        self._add_holiday("Liberty Day", NOV, 1)

        # Christmas Second Day
        self._add_christmas_day_two("Christmas Second Day")

    def _add_subdiv_vt_holidays(self):
        # Town Meeting Day
        if self._year >= 1800:
            self._add_holiday("Town Meeting Day", self._get_nth_weekday_of_month(1, TUE, MAR))

        # Bennington Battle Day
        if self._year >= 1778:
            self._add_observed(self._add_holiday("Bennington Battle Day", AUG, 16))

    def _add_subdiv_wa_holidays(self):
        pass

    def _add_subdiv_wi_holidays(self):
        # Susan B. Anthony Day
        if self._year >= 1976:
            self._add_holiday("Susan B. Anthony Day", FEB, 15)

        if self._year >= 2012:
            # Christmas Eve
            self._add_christmas_eve_holiday()

            # New Year's Eve
            self._add_observed(self._add_new_years_eve("New Year's Eve"), after=False)

    def _add_subdiv_wv_holidays(self):
        # West Virginia Day
        if self._year >= 1927:
            self._add_observed(self._add_holiday("West Virginia Day", JUN, 20))

        # Election Day
        if self._year >= 2008 and self._year % 2 == 0:
            self._add_holiday(
                "Election Day",
                self._get_nth_weekday_of_month(1, MON, NOV) + td(days=+1),
            )

        # Day After Thanksgiving
        if self._year >= 1975:
            self._add_holiday(
                "Day After Thanksgiving",
                self._get_nth_weekday_of_month(4, THU, NOV) + td(days=+1),
            )

    def _add_subdiv_wy_holidays(self):
        pass


class US(UnitedStates):
    pass


class USA(UnitedStates):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td

from holidays.calendars.gregorian import FEB, MAR, APR, SEP, NOV, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Angola(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://www.officeholidays.com/countries/angola/
    https://www.timeanddate.com/holidays/angola/
    """

    country = "AO"

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_observed_holiday(self, dt, before: bool = True) -> None:
        # As of 1995/1/1, whenever a public holiday falls on a Sunday,
        # it rolls over to the following Monday
        # Since 2018 when a public holiday falls on the Tuesday or Thursday
        # the Monday or Friday is also a holiday
        if self.observed and self._year >= 1995:
            for name in self.get_list(dt):
                if self._year <= 2017:
                    if self._is_sunday(dt):
                        self._add_holiday("%s (Observed)" % name, dt + td(days=+1))
                else:
                    if self._is_tuesday(dt) and before:
                        self._add_holiday("%s (Day off)" % name, dt + td(days=-1))
                    elif self._is_thursday(dt):
                        self._add_holiday("%s (Day off)" % name, dt + td(days=+1))

    def _populate(self, year: int) -> None:
        # Observed since 1975
        # TODO do more research on history of Angolan holidays
        if year <= 1974:
            return None

        super()._populate(year)

        # New Year's Day.
        self._add_observed_holiday(self._add_new_years_day("Ano novo"), before=False)
        # Since 2018, if the following year's New Year's Day falls on a
        # Tuesday, the 31st of the current year is also a holiday.
        if self.observed and self._is_monday(DEC, 31) and year >= 2018:
            self._add_holiday("Ano novo (Day off)", DEC, 31)

        # Good Friday.
        self._add_good_friday("Sexta-feira Santa")

        # Carnival.
        self._add_observed_holiday(self._add_carnival_tuesday("Carnaval"))

        # Liberation Movement Day.
        self._add_observed_holiday(self._add_holiday("Dia do Início da Luta Armada", FEB, 4))

        # Day off for International Woman's Day.
        self._add_observed_holiday(self._add_womens_day("Dia Internacional da Mulher"))

        # Southern Africa Liberation Day.
        if year >= 2019:
            self._add_observed_holiday(
                self._add_holiday("Dia da Libertação da África Austral", MAR, 23)
            )

        # Peace Day.
        self._add_observed_holiday(self._add_holiday("Dia da Paz e Reconciliação", APR, 4))

        # May Day.
        self._add_observed_holiday(self._add_labor_day("Dia Mundial do Trabalho"))

        # National Hero Day.
        if year >= 1980:
            self._add_observed_holiday(self._add_holiday("Dia do Herói Nacional", SEP, 17))

        # All Souls' Day.
        self._add_observed_holiday(self._add_all_souls_day("Dia dos Finados"))

        # Independence Day.
        self._add_observed_holiday(self._add_holiday("Dia da Independência", NOV, 11))

        # Christmas Day.
        self._add_observed_holiday(self._add_christmas_day("Dia de Natal e da Família"))


class AO(Angola):
    pass


class AGO(Angola):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td
from datetime import date

from holidays.calendars.gregorian import (
    JAN,
    FEB,
    MAR,
    APR,
    MAY,
    JUN,
    JUL,
    AUG,
    SEP,
    OCT,
    NOV,
    DEC,
    FRI,
    SUN,
)
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Brazil(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    References:
    - https://pt.wikipedia.org/wiki/Feriados_no_Brasil
    - Decreto n. 155-B, de 14.01.1890:
        https://www2.camara.leg.br/legin/fed/decret/1824-1899/decreto-155-b-14-janeiro-1890-517534-publicacaooriginal-1-pe.html
    - Decreto n. 19.488, de 15.12.1930:
        https://www2.camara.leg.br/legin/fed/decret/1930-1939/decreto-19488-15-dezembro-1930-508040-republicacao-85201-pe.html
    """

    country = "BR"
    subdivisions = (
        "AC",  # Acre
        "AL",  # Alagoas
        "AM",  # Amazonas
        "AP",  # Amapá
        "BA",  # Bahia
        "CE",  # Ceará
        "DF",  # Distrito Federal
        "ES",  # Espírito Santo
        "GO",  # Goiás
        "MA",  # Maranhão
        "MG",  # Minas Gerais
        "MS",  # Mato Grosso do Sul
        "MT",  # Mato Grosso
        "PA",  # Pará
        "PB",  # Paraíba
        "PE",  # Pernambuco
        "PI",  # Piauí
        "PR",  # Paraná
        "RJ",  # Rio de Janeiro
        "RN",  # Rio Grande do Norte
        "RO",  # Rondônia
        "RR",  # Roraima
        "RS",  # Rio Grande do Sul
        "SC",  # Santa Catarina
        "SE",  # Sergipe
        "SP",  # São Paulo
        "TO",  # Tocantins
    )

    def __init__(self, *args, **kwargs) -> None:
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        # Decreto n. 155-B, de 14.01.1890
        if year <= 1889:
            return None
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day("Confraternização Universal")

        if 1892 <= year <= 1930:
            # Republic Constitution Day.
            self._add_holiday("Constituição da Republica", FEB, 24)

        # Good Friday.
        self._add_good_friday("Sexta-feira Santa")

        if year not in {1931, 1932}:
            # Tiradentes' Day.
            self._add_holiday("Tiradentes", APR, 21)

        if year >= 1925:
            # Labor Day.
            self._add_labor_day("Dia do Trabalhador")

        if year <= 1930:
            # Discovery of Brazil.
            self._add_holiday("Descobrimento do Brasil", MAY, 3)

            # Abolition of slavery in Brazil.
            self._add_holiday("Abolição da escravidão no Brasil", MAY, 13)

            # Freedom and Independence of American Peoples.
            self._add_holiday("Liberdade e Independência dos Povos Americanos", JUL, 14)

        # Independence Day.
        self._add_holiday("Independência do Brasil", SEP, 7)

        if year <= 1930 or year >= 1980:
            # Our Lady of Aparecida.
            self._add_holiday("Nossa Senhora Aparecida", OCT, 12)

        # All Souls' Day.
        self._add_all_souls_day("Finados")

        # Republic Proclamation Day.
        self._add_holiday("Proclamação da República", NOV, 15)

        if year >= 1922:
            # Christmas Day.
            self._add_christmas_day("Natal")

        # Optional holidays

        # Carnival.
        self._add_carnival_monday("Carnaval")
        self._add_carnival_tuesday("Carnaval")

        # Ash Wednesday.
        self._add_ash_wednesday("Início da Quaresma")

        # Corpus Christi.
        self._add_corpus_christi_day("Corpus Christi")

        # Public Servant's Day.
        self._add_holiday("Dia do Servidor Público", OCT, 28)

        # Christmas Eve.
        self._add_christmas_eve("Véspera de Natal")

        # New Year's Eve.
        self._add_new_years_eve("Véspera de Ano-Novo")

    def _add_subdiv_holidays(self):
        # Lei n. 9.093, de 12.09.1995
        if self._year >= 1996:
            super()._add_subdiv_holidays()

    def _add_subdiv_ac_holidays(self):
        def get_movable_acre(*args) -> date:
            dt = date(self._year, *args)
            if self._year >= 2009 and (
                self._is_tuesday(dt) or self._is_wednesday(dt) or self._is_thursday(dt)
            ):
                dt = self._get_nth_weekday_from(1, FRI, dt)
            return dt

        if self._year >= 2005:
            # Evangelical Day.
            self._add_holiday("Dia do Evangélico", get_movable_acre(JAN, 23))

        if self._year >= 2002:
            # International Women's Day.
            self._add_holiday("Dia Internacional da Mulher", get_movable_acre(MAR, 8))

        # Founding of Acre.
        self._add_holiday("Aniversário do Acre", JUN, 15)

        if self._year >= 2004:
            # Amazonia Day.
            self._add_holiday("Dia da Amazônia", get_movable_acre(SEP, 5))

        # Signing of the Petropolis Treaty.
        self._add_holiday("Assinatura do Tratado de Petrópolis", get_movable_acre(NOV, 17))

    def _add_subdiv_al_holidays(self):
        # Saint John's Day.
        self._add_saint_johns_day("São João")

        # Saint Peter's Day.
        self._add_saints_peter_and_paul_day("São Pedro")

        # Political Emancipation of Alagoas.
        self._add_holiday("Emancipação Política de Alagoas", SEP, 16)

        # Black Awareness Day.
        self._add_holiday("Consciência Negra", NOV, 20)

        if self._year >= 2013:
            self._add_holiday("Dia do Evangélico", NOV, 30)

    def _add_subdiv_am_holidays(self):
        # Elevation of Amazonas to province.
        self._add_holiday("Elevação do Amazonas à categoria de província", SEP, 5)

        if self._year >= 2010:
            self._add_holiday("Consciência Negra", NOV, 20)

    def _add_subdiv_ap_holidays(self):
        if self._year >= 2003:
            # Saint Joseph's Day.
            self._add_saint_josephs_day("São José")

        if self._year >= 2012:
            # Saint James' Day.
            self._add_saint_james_day("São Tiago")

        # Creation of the Federal Territory.
        self._add_holiday("Criação do Território Federal", SEP, 13)

        if self._year >= 2008:
            self._add_holiday("Consciência Negra", NOV, 20)

    def _add_subdiv_ba_holidays(self):
        # Bahia Independence Day.
        self._add_holiday("Independência da Bahia", JUL, 2)

    def _add_subdiv_ce_holidays(self):
        self._add_saint_josephs_day("São José")

        # Abolition of slavery in Ceará.
        self._add_holiday("Abolição da escravidão no Ceará", MAR, 25)

        if self._year >= 2004:
            # Our Lady of Assumption.
            self._add_assumption_of_mary_day("Nossa Senhora da Assunção")

    def _add_subdiv_df_holidays(self):
        # Founding of Brasilia.
        self._add_holiday("Fundação de Brasília", APR, 21)

        self._add_holiday("Dia do Evangélico", NOV, 30)

    def _add_subdiv_es_holidays(self):
        if self._year >= 2020:
            # Our Lady of Penha.
            self._add_holiday("Nossa Senhora da Penha", self._easter_sunday + td(days=+8))

    def _add_subdiv_go_holidays(self):
        # Foundation of Goiás city.
        self._add_holiday("Fundação da cidade de Goiás", JUL, 26)

        # Foundation of Goiânia.
        self._add_holiday("Pedra fundamental de Goiânia", OCT, 24)

    def _add_subdiv_ma_holidays(self):
        # Maranhão joining to independence of Brazil.
        self._add_holiday("Adesão do Maranhão à independência do Brasil", JUL, 28)

    def _add_subdiv_mg_holidays(self):
        # Tiradentes' Execution.
        self._add_holiday("Execução de Tiradentes", APR, 21)

    def _add_subdiv_ms_holidays(self):
        # State Creation Day.
        self._add_holiday("Criação do Estado", OCT, 11)

    def _add_subdiv_mt_holidays(self):
        if self._year >= 2003:
            self._add_holiday("Consciência Negra", NOV, 20)

    def _add_subdiv_pa_holidays(self):
        # Grão-Pará joining to independence of Brazil.
        self._add_holiday("Adesão do Grão-Pará à independência do Brasil", AUG, 15)

    def _add_subdiv_pb_holidays(self):
        # State Founding Day.
        self._add_holiday("Fundação do Estado", AUG, 5)

    def _add_subdiv_pe_holidays(self):
        if self._year >= 2008:
            self._add_holiday(
                # Pernambuco Revolution.
                "Revolução Pernambucana",
                self._get_nth_weekday_of_month(1, SUN, MAR),
            )

    def _add_subdiv_pi_holidays(self):
        # Piauí Day.
        self._add_holiday("Dia do Piauí", OCT, 19)

    def _add_subdiv_pr_holidays(self):
        # Emancipation of Paraná.
        self._add_holiday("Emancipação do Paraná", DEC, 19)

    def _add_subdiv_rj_holidays(self):
        if self._year >= 2008:
            # Saint George's Day.
            self._add_saint_georges_day("São Jorge")

        if self._year >= 2002:
            self._add_holiday("Consciência Negra", NOV, 20)

    def _add_subdiv_rn_holidays(self):
        if self._year >= 2000:
            # Rio Grande do Norte Day.
            self._add_holiday("Dia do Rio Grande do Norte", AUG, 7)

        if self._year >= 2007:
            # Uruaçú and Cunhaú Martyrs Day.
            self._add_holiday("Mártires de Cunhaú e Uruaçuu", OCT, 3)

    def _add_subdiv_ro_holidays(self):
        self._add_holiday("Criação do Estado", JAN, 4)

        if self._year >= 2002:
            self._add_holiday("Dia do Evangélico", JUN, 18)

    def _add_subdiv_rr_holidays(self):
        self._add_holiday("Criação do Estado", OCT, 5)

    def _add_subdiv_rs_holidays(self):
        # Gaucho Day.
        self._add_holiday("Dia do Gaúcho", SEP, 20)

    def _add_subdiv_sc_holidays(self):
        if self._year >= 2004:
            dt = date(self._year, AUG, 11)
            if self._year >= 2005:
                dt = self._get_nth_weekday_from(1, SUN, dt)
            # Santa Catarina State Day.
            self._add_holiday("Dia do Estado de Santa Catarina", dt)

        dt = date(self._year, NOV, 25)
        if 1999 <= self._year != 2004:
            dt = self._get_nth_weekday_from(1, SUN, dt)
        # Saint Catherine of Alexandria Day.
        self._add_holiday("Dia de Santa Catarina de Alexandria", dt)

    def _add_subdiv_se_holidays(self):
        # Sergipe Political Emancipation Day.
        self._add_holiday("Emancipação política de Sergipe", JUL, 8)

    def _add_subdiv_sp_holidays(self):
        if self._year >= 1997:
            # Constitutionalist Revolution.
            self._add_holiday("Revolução Constitucionalista", JUL, 9)

    def _add_subdiv_to_holidays(self):
        if self._year >= 1998:
            # Autonomy Day.
            self._add_holiday("Dia da Autonomia", MAR, 18)

        # Our Lady of Nativity.
        self._add_nativity_of_mary_day("Nossa Senhora da Natividade")

        self._add_holiday("Criação do Estado", OCT, 5)


class BR(Brazil):
    pass


class BRA(Brazil):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.calendars.gregorian import JUL, AUG, SEP
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Nicaragua(HolidayBase, ChristianHolidays, InternationalHolidays):
    country = "NI"
    default_language = "es"
    subdivisions = (
        "AN",
        "AS",
        "BO",
        "CA",
        "CI",
        "CO",
        "ES",
        "GR",
        "JI",
        "LE",
        "MD",
        "MN",
        "MS",
        "MT",
        "NS",
        "RI",
        "SJ",
    )
    supported_languages = ("en_US", "es", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        # Default subdivision to MN; prov for backwards compatibility
        if not kwargs.get("subdiv", kwargs.get("prov")):
            kwargs["subdiv"] = "MN"
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("Año Nuevo"))

        # Maundy Thursday.
        self._add_holy_thursday(tr("Jueves Santo"))

        # Good Friday.
        self._add_good_friday(tr("Viernes Santo"))

        # Labor Day.
        self._add_labor_day(tr("Día del Trabajo"))

        if year >= 1979:
            # Revolution Day.
            self._add_holiday(tr("Día de la Revolución"), JUL, 19)

        # Battle of San Jacinto Day.
        self._add_holiday(tr("Batalla de San Jacinto"), SEP, 14)

        # Independence Day.
        self._add_holiday(tr("Día de la Independencia"), SEP, 15)

        # Virgin's Day.
        self._add_immaculate_conception_day(tr("Concepción de María"))

        # Christmas.
        self._add_christmas_day(tr("Navidad"))

    def _add_subdiv_mn_holidays(self):
        # Descent of Saint Dominic.
        self._add_holiday(tr("Bajada de Santo Domingo"), AUG, 1)

        # Ascent of Saint Dominic.
        self._add_holiday(tr("Subida de Santo Domingo"), AUG, 10)


class NI(Nicaragua):
    pass


class NIC(Nicaragua):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.calendars.gregorian import FEB, MAR, JUN, SEP, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Malta(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://www.gov.mt/en/About%20Malta/Pages/Public%20Holidays.aspx

    [Att 10 tal-1980]
        Oldest Maltese Holidays Law available online in full.
        https://legislation.mt/eli/act/1980/10/mlt
    [A.L. 40 tal-1987]
        Additional Holidays added.
        https://legislation.mt/eli/ln/1987/8/mlt
    [Att 8 tal-1989]
        Additional Holidays added.
        https://legislation.mt/eli/act/1989/8
    [Att 2 tal-2005]
        If fall on weekends then not observed in terms of vacation leave.
        https://legislation.mt/eli/act/2005/2/eng
    [Att 4 tal-2021]
        Revert Act II of 2005 changes for vacation leave.
        https://legislation.mt/eli/cap/252/20210212/mlt
    """

    country = "MT"
    default_language = "mt"
    supported_languages = ("en_MT", "mt")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        # Earliest available source is 1980
        if year <= 1979:
            return None
        super()._populate(year)

        # L-Ewwel tas-Sena
        # Status: In-Use.

        # New Year's Day
        self._add_new_years_day(tr("L-Ewwel tas-Sena"))

        # Il-Festa tan-Nawfraġju ta' San Pawl
        # Status: In-Use.
        # Started in 1987 via Act LX of 1987.

        if year >= 1987:
            # Feast of St. Paul's Shipwreck
            self._add_holiday(tr("Il-Festa tan-Nawfraġju ta' San Pawl"), FEB, 10)

        # Il-Festa ta' San Ġużepp
        # Status: In-Use.
        # Started in 1987 via Act LX of 1987.

        if year >= 1987:
            # Feast of St. Joseph
            self._add_saint_josephs_day(tr("Il-Festa ta' San Ġużepp"))

        # Jum il-Ħelsien
        # Status: In-Use.
        # Started in 1980 Act X of 1980.
        # Not presented in 1987-1988

        if year <= 1986 or year >= 1989:
            # Freedom Day
            self._add_holiday(tr("Jum il-Ħelsien"), MAR, 31)

        # Il-Ġimgħa l-Kbira
        # Status: In-Use.

        # Good Friday
        self._add_good_friday(tr("Il-Ġimgħa l-Kbira"))

        # Jum il-Ħaddiem
        # Status: In-Use.

        # Worker's Day
        self._add_labor_day(tr("Jum il-Ħaddiem"))

        # Sette Giugno
        # Status: In-Use.
        # Start in 1989 via Act VIII of 1989.

        if year >= 1989:
            # Sette Giugno
            self._add_holiday(tr("Sette Giugno"), JUN, 7)

        # Il-Festa ta' San Pietru u San Pawl
        # Status: In-Use.
        # Started in 1987 via Act LX of 1987.

        if year >= 1987:
            # Feast of St. Peter and St. Paul
            self._add_saints_peter_and_paul_day(tr("Il-Festa ta' San Pietru u San Pawl"))

        # Il-Festa ta' Santa Marija
        # Status: In-Use.

        # Feast of the Assumption
        self._add_assumption_of_mary_day(tr("Il-Festa ta' Santa Marija"))

        # Jum il-Vitorja
        # Status: In-Use.
        # Started in 1987 via Act LX of 1987.
        # While this concides with Nativity Of Mary Day, the two are considered separate.

        if year >= 1987:
            # Feast of Our Lady of Victories
            self._add_holiday(tr("Jum il-Vitorja"), SEP, 8)

        # Jum l-Indipendenza
        # Status: In-Use.
        # Started in 1987 via Act LX of 1987.

        if year >= 1987:
            # Independence Day
            self._add_holiday(tr("Jum l-Indipendenza"), SEP, 21)

        # Il-Festa tal-Immakulata Kunċizzjoni
        # Status: In-Use.
        # Started in 1987 via Act LX of 1987.

        if year >= 1987:
            # Feast of the Immaculate Conception
            self._add_immaculate_conception_day(tr("Il-Festa tal-Immakulata Kunċizzjoni"))

        # Jum ir-Repubblika
        # Status: In-Use.

        # Republic Day
        self._add_holiday(tr("Jum ir-Repubblika"), DEC, 13)

        # Il-Milied
        # Status: In-Use.

        # Christmas Day
        self._add_christmas_day(tr("Il-Milied"))


class MT(Malta):
    pass


class MLT(Malta):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from holidays.calendars import _CustomIslamicCalendar
from holidays.calendars.gregorian import JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, NOV, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays, IslamicHolidays


class Gabon(HolidayBase, ChristianHolidays, InternationalHolidays, IslamicHolidays):
    """
    References:
      - https://en.wikipedia.org/wiki/Public_holidays_in_Gabon
      - https://www.timeanddate.com/holidays/gabon
      - https://www.officeholidays.com/countries/gabon
      - http://www.travail.gouv.ga/402-evenements/489-liste-des-jours-feries/
    """

    country = "GA"

    def __init__(self, *args, **kwargs) -> None:
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self, calendar=GabonIslamicCalendar())
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        # On 17 August 1960, Gabon gained independence from France.
        if year <= 1960:
            return None

        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day("New Year's Day")

        # Easter Monday.
        self._add_easter_monday("Easter Monday")

        # Women's Rights Day.
        if year >= 2015:
            self._add_holiday("Women's Rights Day", APR, 17)

        # Labour Day.
        self._add_labor_day("Labour Day")

        # Ascension Day.
        self._add_ascension_thursday("Ascension Day")

        # Whit Monday.
        self._add_whit_monday("Whit Monday")

        # Assumption Day.
        self._add_assumption_of_mary_day("Assumption Day")

        # Independence Day.
        self._add_holiday("Independence Day", AUG, 16)
        self._add_holiday("Independence Day Holiday", AUG, 17)

        # All Saints' Day.
        self._add_all_saints_day("All Saints' Day")

        # Christmas Day.
        self._add_christmas_day("Christmas Day")

        # Eid al-Fitr.
        self._add_eid_al_fitr_day("Eid al-Fitr")

        # Eid al-Adha.
        self._add_eid_al_adha_day("Eid al-Adha")


class GA(Gabon):
    pass


class GAB(Gabon):
    pass


class GabonIslamicCalendar(_CustomIslamicCalendar):
    EID_AL_ADHA_DATES = {
        2001: (MAR, 6),
        2002: (FEB, 23),
        2003: (FEB, 12),
        2004: (FEB, 2),
        2005: (JAN, 21),
        2006: ((JAN, 10), (DEC, 31)),
        2007: (DEC, 20),
        2008: (DEC, 9),
        2009: (NOV, 28),
        2010: (NOV, 17),
        2011: (NOV, 7),
        2012: (OCT, 26),
        2013: (OCT, 15),
        2014: (OCT, 5),
        2015: (SEP, 24),
        2016: (SEP, 13),
        2017: (SEP, 2),
        2018: (AUG, 22),
        2019: (AUG, 11),
        2020: (JUL, 31),
        2021: (JUL, 20),
        2022: (JUL, 9),
        2023: (JUN, 28),
    }

    EID_AL_FITR_DATES = {
        2001: (DEC, 17),
        2002: (DEC, 6),
        2003: (NOV, 26),
        2004: (NOV, 14),
        2005: (NOV, 4),
        2006: (OCT, 24),
        2007: (OCT, 13),
        2008: (OCT, 2),
        2009: (SEP, 21),
        2010: (SEP, 10),
        2011: (AUG, 31),
        2012: (AUG, 19),
        2013: (AUG, 8),
        2014: (JUL, 29),
        2015: (JUL, 18),
        2016: (JUL, 7),
        2017: (JUN, 26),
        2018: (JUN, 15),
        2019: (JUN, 4),
        2020: (MAY, 24),
        2021: (MAY, 13),
        2022: (MAY, 2),
        2023: (APR, 21),
    }



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td

from holidays.calendars.gregorian import (
    JAN,
    MAR,
    APR,
    MAY,
    JUN,
    JUL,
    AUG,
    SEP,
    OCT,
    NOV,
    DEC,
    MON,
    FRI,
)
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class SouthAfrica(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    http://www.gov.za/about-sa/public-holidays
    https://en.wikipedia.org/wiki/Public_holidays_in_South_Africa
    """

    country = "ZA"
    special_holidays = {
        1999: (
            (JUN, 2, "National and provincial government elections"),
            (DEC, 31, "Y2K changeover"),
        ),
        2000: (JAN, 2, "Y2K changeover"),
        2004: (APR, 14, "National and provincial government elections"),
        2006: (MAR, 1, "Local government elections"),
        2008: (MAY, 2, "Public holiday by presidential decree"),
        2009: (APR, 22, "National and provincial government elections"),
        2011: (
            (MAY, 18, "Local government elections"),
            (DEC, 27, "Public holiday by presidential decree"),
        ),
        2014: (MAY, 7, "National and provincial government elections"),
        2016: (
            (AUG, 3, "Local government elections"),
            (DEC, 27, "Public holiday by presidential decree"),
        ),
        2019: (MAY, 8, "National and provincial government elections"),
        2021: (NOV, 1, "Municipal elections"),
        2022: (DEC, 27, "Public holiday by presidential decree"),
    }

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_observed(self, dt: date) -> None:
        # As of 1995/1/1, whenever a public holiday falls on a Sunday,
        # it rolls over to the following Monday
        if self.observed and self._is_sunday(dt) and self._year >= 1995:
            self._add_holiday("%s (Observed)" % self[dt], dt + td(days=+1))

    def _populate(self, year):
        # Observed since 1910, with a few name changes
        if year <= 1909:
            return None

        super()._populate(year)

        self._add_observed(self._add_new_years_day("New Year's Day"))

        self._add_good_friday("Good Friday")

        self._add_easter_monday("Family Day" if year >= 1980 else "Easter Monday")

        if year <= 1951:
            name = "Dingaan's Day"
        elif year <= 1979:
            name = "Day of the Covenant"
        elif year <= 1994:
            name = "Day of the Vow"
        else:
            name = "Day of Reconciliation"
        self._add_observed(self._add_holiday(name, DEC, 16))

        self._add_christmas_day("Christmas Day")

        self._add_observed(
            self._add_christmas_day_two("Day of Goodwill" if year >= 1980 else "Boxing Day")
        )

        if year >= 1995:
            self._add_observed(self._add_holiday("Human Rights Day", MAR, 21))

            self._add_observed(self._add_holiday("Freedom Day", APR, 27))

            self._add_observed(self._add_labor_day("Workers' Day"))

            self._add_observed(self._add_holiday("Youth Day", JUN, 16))

            self._add_observed(self._add_holiday("National Women's Day", AUG, 9))

            self._add_observed(self._add_holiday("Heritage Day", SEP, 24))

        # Special holiday http://tiny.cc/za_y2k
        if self.observed and year == 2000:
            self._add_holiday("Y2K changeover (Observed)", JAN, 3)

        # Historic public holidays no longer observed
        if 1952 <= year <= 1973:
            self._add_holiday("Van Riebeeck's Day", APR, 6)
        elif 1980 <= year <= 1994:
            self._add_holiday("Founder's Day", APR, 6)

        if 1987 <= year <= 1989:
            self._add_holiday("Workers' Day", self._get_nth_weekday_of_month(1, FRI, MAY))

        if year <= 1993:
            self._add_ascension_thursday("Ascension Day")

        if year <= 1951:
            self._add_holiday("Empire Day", MAY, 24)

        if year <= 1960:
            self._add_holiday("Union Day", MAY, 31)
        elif year <= 1993:
            self._add_holiday("Republic Day", MAY, 31)

        if 1952 <= year <= 1960:
            self._add_holiday("Queen's Birthday", self._get_nth_weekday_of_month(2, MON, JUL))

        if 1961 <= year <= 1973:
            self._add_holiday("Family Day", JUL, 10)

        if year <= 1951:
            self._add_holiday("King's Birthday", self._get_nth_weekday_of_month(1, MON, AUG))

        if 1952 <= year <= 1979:
            self._add_holiday("Settlers' Day", self._get_nth_weekday_of_month(1, MON, SEP))

        if 1952 <= year <= 1993:
            self._add_holiday("Kruger Day", OCT, 10)


class ZA(SouthAfrica):
    pass


class ZAF(SouthAfrica):
    pass
